require 'test/unit'

require 'crdb/cache'

# This class is so we can mixin the Cache module and test it.
class CacheClass
  include CRDB::Cache
end


class TC_TestCache < Test::Unit::TestCase

  def make_cacheclass_object
    cache = CacheClass.new()
    cache.cache_dir = "/tmp/rubycache/testing"
    cache.cache_namespace = "TC_TestCache"

    return cache
  end

  # The ORDER the tests run is important!

  # TEST: Create a CacheClass object
  def test_00_cache_new
    cache = make_cacheclass_object
    assert_equal(cache.class.to_s, 'CacheClass')
  end

  # TEST: Create a CacheClass object without a cache dir which should 
  # throw an error when we turn caching on.
  def test_01_no_cache_dir
    cache = make_cacheclass_object
    cache = CacheClass.new()
    cache.cache_namespace = "TC_TestCache"

    assert_raise(RuntimeError) { cache.caching_on() }
  end

  # TEST: Create a CacheClass object without a cache namespace which should 
  # throw an error when we turn caching on.
  def test_02_no_cache_namespace
    cache = make_cacheclass_object
    cache = CacheClass.new()
    cache.cache_dir = "/tmp/rubycache/testing"

    assert_raise(RuntimeError) { cache.caching_on() }
  end

  # TEST: Create a CacheClass object and attempt to cache something.
  def test_03_simple_write_read
    c = make_cacheclass_object
    cache = c.caching_on()
    cache.write("key", "value")
    value = cache.read("key")
    assert_equal(value, "value")  
  end

  # TEST: Create a CacheClass object, write something, then delete it.
  def test_04_simple_delete
    c = make_cacheclass_object
    cache = c.caching_on()
    cache.write("key", "value")
    value = cache.read("key")
    cache.delete("key")
    value = cache.read("key")
    assert_nil(value)  
  end

  # TEST: Create a CacheClass object, and read something that should have 
  # been written in an earlier run.
  def test_05_read_old
    c = make_cacheclass_object
    cache = c.caching_on()
    value_old = cache.read("key_old")
    assert_equal(value_old, "value_old")
  end

  # TEST: Create a CacheClass object, write two objects, then clear entire
  # cache.
  def test_06_simple_delete
    c = make_cacheclass_object
    cache = c.caching_on()
    cache.write("key1", "value1")
    cache.write("key2", "value2")
    value1 = cache.read("key1")
    value2 = cache.read("key2")
    assert_equal(value1, "value1")
    assert_equal(value2, "value2")  
    c.clear()
    value1 = cache.read("key1")
    value2 = cache.read("key2")
    assert_nil(value1)
    assert_nil(value2)
  end


  # TEST: Create a CacheClass object, and write something that will be 
  # read in a later run.
  def test_99_read_old
    c = make_cacheclass_object
    cache = c.caching_on()
    cache.write("key_old", "value_old")
  end


end
