module CRDB
 class Adminunit < ActiveRecord::Base

  require 'crdb/misc'
  require 'crdb/resource'
  require 'crdb/adminownership'

  # The "join table" ownerships links resources to departments
  has_many :adminownerships, :class_name => "CRDB::Adminownership"
  has_many :resources, :class_name => "CRDB::Resource", :through => :adminownerships


  # The code must be four characters conisting of letters and numbers
  validates_presence_of :code
  validates_presence_of :long_name
  validates_uniqueness_of :code
  validate :code_must_be_four

  def code_must_be_four
    if (code.nil? || code.empty?)
      msg = "code cannot be nil or empty"
      errors.add_to_base(msg) 
    elsif (code.length() < 4)
      msg = "code must be exactly four characters long (#{code})"
      errors.add_to_base(msg) 
    elsif (code !~ /^[A-Z][A-Z0-9]{3}$/)
      msg = "code must consist of four characters, each of which is an" +
            "upper-case letter or a number (#{code})"
      errors.add_to_base(msg) 
    end
  end 

  # The URL where the admin codes are kept
  Adminunit_URL = 'http://www.stanford.edu/dept/pres-provost/budget/org/export_sys.asc'

  def to_s
    @id = self.id
    @code = self.code
    @long_name = self.long_name
    return "(id: #{@id}, code: #{@code}, long_name: #{@long_name})" 
  end

  # Given an array of CRDB::Adminunit object, format into a string.
  # Takes these options:
  #   * :separator (defaults to "-").
  def (CRDB::Adminunit).format_multiple_adminunits(array_of_adminunits,  *options_array)

    # Extract any options and process
    separator = "-"
    if (options_array.length > 0)
      options_hash = options_array[0]
      if (options_hash.key?(:separator))
        separator = options_hash[:separator]
      end
    end

    codes = [] 
    array_of_adminunits.each do |adminunit|
      codes.push(adminunit.code)
    end

    if (codes.empty?)
      return "NO_ADMIN_UNITS"
    else
      return codes.join(separator) 
    end

  end

  # Given a comma-separated list of admin unuts (e.g., "CHAR, VZZY") 
  # parse and return an array of CRDB::Adminunit objects, one for each.
  # If no recognizable admin units, returns an empty array.
  def (CRDB::Adminunit).parse_adminunits_string(adminunits_string)
    if ((CRDB::Misc).nilorempty?(adminunits_string))
      # If the string is empty or nil, return an empty array.
      return [] 
    else
      adminunit_codes = adminunits_string.split(",") 
     
      # Store the resulting CRDB::Adminunit objects in the array admin_units.
      admin_units = [] 

      adminunit_codes.each do |adminunit_code|
        # Strip the code
        adminunit_code.strip!

        # See if there is a CRDB::Adminunit with this code.
        adminunit = (CRDB::Adminunit).find(:first, :conditions => {:code => adminunit_code})

        if (!adminunit.nil?)
          # We found one, so add it the results array.
          admin_units.push(adminunit)
        end
      end
    end

    return admin_units
  end


   # Populate an array of CRDB::Adminunit objects with the data stored 
   # at (CRDB::Adminunit)_URL. 
   def (CRDB::Adminunit).get_adminunits_from_url
     require 'net/http'
     http = Net::HTTP.new('www.stanford.edu', 80)
     response, data = http.get(Adminunit_URL)

     adminunits = [] 

     # Parse the data loading the information into the database
     lines = data.split("\n")

     lines.each do |line|

       # The lines are tab-delimited. The first field is the
       # four-character admin unit code. The sixth field is the "Display
       # Description" field. 
       # 
       # Here is an example line from the web page: 
       ##########
       # ABIQ    AAAA    Public Affairs  Public Affairs  Public Affairs  Public Affairs  FALSE   2/4/2003        Office of Public Affairs        FALSE                           ABIQ    AAAA    AA00            
       ##########
       fields = line.split("\t")

       code      = fields[0]
       long_name = fields[5]

       # If either code or long_name is empty, skip
       if ((CRDB::Misc).nilorempty?(code) || (CRDB::Misc).nilorempty?(long_name))
         next 
       end 

       code.strip!
       long_name.strip!

      # Remove any double quotes on long_name
      long_name.gsub!(/\"/, '')

      # Skip the first line where code is "10 Char Name"
      if (code =~ /10 Char Name/)
        next 
      end 

      a = (CRDB::Adminunit).new
      a.code = code
      a.long_name = long_name
 
      adminunits.push(a)
    end

    return adminunits
  end

  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###
  # "Mix-in" the Fake module
  require 'crdb/fake.rb'
  include CRDB::Fake

  @@fake_method_to_value_hash = nil
  def (CRDB::Adminunit).initialize_fake_method_to_value_hash()
    fake_codes = ['ABCD', 'AA99']
    fake_long_names = ['Pet Grooming Administration', 'Music and Holistic Healing']

    @@fake_method_to_value_hash = 
    {
      'code'      => fake_codes,
      'long_name' => fake_long_names,
    }
  end

  def (CRDB::Adminunit).get_fake_method_to_value_hash()
    return @@fake_method_to_value_hash
  end

  def (CRDB::Adminunit).delete_all_fakes()
    (CRDB::Adminunit).initialize_fake_method_to_value_hash()
    fake_long_names = @@fake_method_to_value_hash['long_name']
    (CRDB::Adminunit).delete_all(["long_name = ?", fake_long_names[0]])
    (CRDB::Adminunit).delete_all(["long_name = ?", fake_long_names[1]])
  end   
  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###

 end # matches "class Adminunit"
end # matches "module CRDB"

