
# The Building object 
# QB::         The quad-building number
# short_name:: The building's short name (e.g., Tressider)
# full_name:: The building's short name (e.g., Tressider)
# very_short_name:: A version of the building name short enough to fit into Zimbra's building field.

module CRDB
 class Building < ActiveRecord::Base

  has_one :resource, :class_name => "CRDB::Resource"

  ## VALIDATIONS
  validates_presence_of :QB
  validates_presence_of :full_name
  validates_presence_of :short_name
  validates_presence_of :very_short_name
  validates_uniqueness_of :QB
  validates_uniqueness_of :full_name
  validates_uniqueness_of :short_name
  validates_uniqueness_of :very_short_name

  validates_length_of :very_short_name, :maximum=>32

  validate :validate_QB_format
  # An ActiveRecord validation to ensure that the QB field is correctly formatted.
  def validate_QB_format
    if (self.QB !~ /^\d\d\-\d\d\d[A-Z]?$/)
      errors.add_to_base("the QB must be of the form 'A-B' where 'A' " +
                         "consists of two digits and B " +
                         "is three digits followed by an optional " +
                         "upper-case letter (yours: '#{self.QB}')") 
    end
  end

  def to_s
    id = self.id
    qb = self.QB
    short_name = self.short_name
    full_name = self.full_name
    very_short_name = self.very_short_name
    note = self.note
    return "(id: #{id}, QB: #{qb}, short_name: #{short_name}, " +
           "full_name: #{full_name}, very_short_name: #{very_short_name}), " +
           "notes: #{notes}"
  end

  # Returns just the building's short name.
  def to_s_short
    return self.short_name
  end

  def eql?(building)
    return (
            (self.QB.eql?(building.QB)) && 
            (self.short_name.eql?(building.short_name)) && 
            (self.full_name.eql?(building.full_name))
           )
  end


  # Split QBFR into [QB, FR]
  #   Possible return values: 
  #     Q-B-F-R --> [Q-B, F-R]
  #     Q-B-F   --> [Q-B, F]
  #     Q-B     --> [Q-B, nil]
  #     Q       --> [Q, nil]
  #     empty   --> [nil, nil]
  #     nil     --> [nil, nil]
  def (CRDB::Building).parseQBFR(qbfr)
    if (qbfr.nil? || qbfr.empty?)
      return [nil, nil]
    elsif (qbfr =~ /^(\w+\-\w+)\-(\w+\-\w+)$/)
      return [$1, $2]
    elsif (qbfr =~ /^(\w+\-\w+)-(\w+)$/)
      return [$1, $2]
    elsif (qbfr =~ /^(\w+\-\w+)$/)
      return [$1, nil]
    elsif (qbfr =~ /^(\w+)$/)
      return [$1, nil]
    else
      raise "problem parsing QBFR #{qbfr}"
    end 
  end


  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###
  # "Mix-in" the Fake module
  require 'crdb/fake.rb'
  include CRDB::Fake

  @@fake_method_to_value_hash = nil
  def (CRDB::Building).initialize_fake_method_to_value_hash()
    fake_QBs = ['99-999Z', '99-998Q']
    fake_full_names = ['Alfred E. Newman Memorial Auditorium', 'Lottie & Tootsie Laboratories']
    fake_short_names = ['Alfred Newman Memorial Auditorium', 'Lottie & Tootsie Labs']
    fake_very_short_names = ['Newman Memorial Auditorium', 'Lottie Labs']

    @@fake_method_to_value_hash = 
    {
      'QB'              => fake_QBs,
      'full_name'       => fake_full_names,
      'short_name'      => fake_short_names,
      'very_short_name' => fake_very_short_names,
    }
  end

  def (CRDB::Building).get_fake_method_to_value_hash()
    return @@fake_method_to_value_hash
  end

  def (CRDB::Building).delete_all_fakes()
    (CRDB::Building).initialize_fake_method_to_value_hash()
    fake_full_names = @@fake_method_to_value_hash['full_name']
    (CRDB::Building).delete_all(["full_name = ?", fake_full_names[0]])
    (CRDB::Building).delete_all(["full_name = ?", fake_full_names[1]])
  end   
  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###

 end # matches "class Building"
end # matches "module CRDB"
