require 'tempfile'

module CRDB
 class CRDBRemctl < Object

  require 'crdb/misc'
  require 'crdb/session'
  require 'crdb/resource'

  ### Process options.
  ### Returns a hash with these keys:
  ###   'platform'
  def process_options()
    require 'getoptlong'
    opts = GetoptLong.new(
      [ '--platform',      '-p', GetoptLong::OPTIONAL_ARGUMENT ],
      [ '--sync_commands', '-s', GetoptLong::NO_ARGUMENT ],
      [ '--dry_run',       '-n', GetoptLong::NO_ARGUMENT ]
                     )

    platform      = 'production'
    dry_run       = false
    sync_commands = false

    opts.each do |opt, arg|
      case opt
        when '--platform'
          platform = arg
        when '--dry_run'
          dry_run = true
        when '--sync_commands'
          sync_commands = true
        else
          raise "unrecognized option '#{opt}'"
      end
    end

    begin
      platform =  (CRDB::Misc).arg_process_platform_option(platform)
    rescue
      raise "the -p option should be followed by either 'production' or 'development'"
    end

    return {
            'platform' => platform,
            'dry_run' => dry_run,
            'sync_commands' => sync_commands,
           }
  end

  def platform
    options_hash = process_options()
    return options_hash['platform']
  end


  ### The help command
  def help()
    platform = platform()
    version = (CRDB::Misc).get_crdb_version()

    help_string = <<EOS
NAME
    crdb - Manage Calendar Resource Database

SYNOPSIS
    crdb help
    crdb list (locations|nonlocations|all)
    crdb find <fieldname> <string>
    crdb show <resource_uid>
    crdb table <table_name>
    crdb create [-n] <full_path_to_xml_file>
    crdb set <resource_uid> <fieldname> <string> [PARTLY WORKING]
    crdb DELETE <resource_uid>
    crdb diff [-s] <zimbra_server>
    crdb sync [-n] <resource_uid> <zimbra_server>
    crdb system_check


DESCRIPTION
    This is an interface for Calendar Resource Database management through
    remctl calls.

NOTE
    All commands (except help) understand the "-p" option. This option
    takes an argument of either 'development' or 'production'. For
    example, "crdb find -p development list all" will list all the
    resources in the development database. If omitted, the platform
    is assumed to be production.

COMMANDS
    help
        Display this help screen.

    list locations
    list nonlocations
    list all
        List resource uids for all locations, nonlocations, or both.

    find
    find <fieldname> <string>
        Displays all resources that have <fieldname> containing <string>.
        Use "crdb find" with no other arguments to display the possible
        values for <fieldname>. <string> can contain the wildcard
        character "*". Example: crdb find contact_fullname "smith"

    show <resource_uid>
        Displays the details on the resource with the supplied uid.

    table <table_name>
        Dumps the contents of <table_name>. If <table_name> is omitted,
        will display a list of table names that can be dumped.

    create [-n] <full_path_to_xml_file>
        Uses the XML file at <full_path_to_xml_file> to create a new
        resource in the CRDB. The "-n" option causes a "dry-run" (nothing
        is saved).

    set <resource_uid> <fieldname> <string>
        Sets the field <fieldname> to <string>. If <fieldname> is
        multi-valued, put a "+" or "-" at the start of <value> (e.g.,
        "+CAHR").

    diff [-s] <zimbra_server>
        Displays the difference between the resources in the CRDB and the
        resource in the supplied Zimbra server. If the "-s" option is
        given, all the remctl commands necessary to sync any out-of-sync
        resources are output.

    sync <resource_uid> <zimbra_server>
        Updates the resource attributes for the resource <resource_uid> on
        <zimbra_server> with the values in the CRDB. The "-n" option
        causes a "dry-run" (nothing is saved).

    DELETE <resource_uid>
        Deletes the resource with the supplied uid. WARNING! This delete
        cannot be rolled-back.

    system_check
        Does some basic checks to see if the configuration is set up
        correctly and returns some of the configuration data.

    (CRDB version: #{version})
EOS

    return help_string
  end


  ### The show command
  def show()
    platform = platform()
    connect_info = (CRDB::Session).connect(platform)

    # Get the uid
    if (ARGV.length >= 2)
      uid = ARGV[1]
    else
      raise "missing required resource uid"
    end

    # Load this resource
    all_resources = (CRDB::Resource).find_by_uid(uid)

    if (all_resources.nil?)
      return "no #{platform} resources found with uid '#{uid}'"
    else
      return all_resources.to_s
    end
  end

  ### The list command
  ### Lists resource uid's along with their display names.
  def list(verbose = false)
    platform = platform()
    connect_info = (CRDB::Session).connect(platform)

    # Get the scope
    if (ARGV.length >= 2)
      scope = ARGV[1]
    else
      raise "for 'list' you need to specify one of locations, nonlocations, or all"
    end

    # scope should be one of 'all', 'locations', or 'nonlocations'
    if (scope =~ /\Aall/)
      scope = 'all'
      resources = (CRDB::Resource).all
    elsif (scope =~ /\Aloc/)
      scope = 'locations'
      resources = (CRDB::Resource).includes(:type => :category).where('categories.code' => "lo")
    elsif (scope =~ /\Anon/)
      scope = 'nonlocations'
      resources = (CRDB::Resource).includes(:type => :category).where.not('categories.code' => "lo")
    else
      raise "for 'list' you need to specify one of locations, nonlocations, or all"
    end

    return_value = ''
    resources.each do |resource|
      if (verbose)
        return_value += resource.to_s + "\n"
        return_value += "--------------------"  + "\n"
      else
        return_value += %Q/#{resource.uid} [#{resource.display_name}]\n/
      end
    end

    return return_value
  end

  ### The find command
  def find()
    platform = platform()
    connect_info = (CRDB::Session).connect(platform)

    # If there are no arguments, then show the
    # field list
    if (ARGV.length == 1)
      available_fields = (CRDB::Resource).find_resource_available_fields()
      available_fields_sorted = available_fields.sort
      text = "Available fields to search on are:\n  " + available_fields_sorted.join("\n  ")
      return text
    elsif (ARGV.length == 2)
      raise "for 'find' you need to specify both a field name and string"
    else
      return (CRDB::Resource).find_resource_search_results(ARGV[1], ARGV[2])
    end
  end


  ### The find command
  def dELETE()
    platform = platform()
    connect_info = (CRDB::Session).connect(platform)

    # Get the uid
    if (ARGV.length >= 2)
      uid = ARGV[1]
    else
      raise "missing required resource uid"
    end

    # Step 1. Get the resource
    resource = (CRDB::Resource).find_by_uid(uid)

    if (resource.nil?)
      raise "could not find resource '#{uid}' in '#{platform}'"
    else
      resource.destroy

      new_resource = (CRDB::Resource).find_by_uid(uid)
      if (new_resource.nil?)
        return "resource #{uid} in #{platform} deleted"
      else
        raise "resource #{uid} delete failed"
      end
    end
  end


  ### The sync command
  def sync()
    options_hash = process_options()
    platform = options_hash['platform']
    dry_run  = options_hash['dry_run']

    connect_info = (CRDB::Session).connect(platform)

    # Get the uid and zimbra_server name
    if (ARGV.length >= 3)
      uid = ARGV[1]
      zimbra_server = ARGV[2]
    else
      raise "missing required resource uid or zimbra server"
    end

    # Step 1. Get the resource
    resource = (CRDB::Resource).find_by_uid(uid)

    if (resource.nil?)
      raise "could not find resource '#{uid}' in '#{platform}'"
    else
      # Make a CRDB::Zimbra object. Since this CRDB::Zimbra object needs to call
      # a remctl, we need to add a kerberos object.
      require 'crdb/zimbra'
      zimbra = (CRDB::Zimbra).new()
      zimbra.server = zimbra_server
      zimbra.kerberos = (CRDB::CRDBRemctl).make_kerberos()
      zimbra.sync_resource(resource, nil, {:dry_run => dry_run})
    end
  end

  ### The diff command
  def diff()
    options_hash = process_options()

    platform      = options_hash['platform']
    sync_commands = options_hash['sync_commands']

    connect_info = (CRDB::Session).connect(platform)

    # Get the uid and zimbra_server name
    if (ARGV.length >= 2)
      zimbra_server = ARGV[1]
    else
      raise "missing required zimbra server argument"
    end

    # Step 1. Create a Kerberos ticket object
    kerb = (CRDB::CRDBRemctl).make_kerberos()

    # Step 1. Get the difference arrays
    in_zimbra_only, in_database_only, in_both =
      (CRDB::Resource).find_differences(zimbra_server, {:caching => false,
                                                :kerberos => kerb})

    # Step 2. If we are showing the sync commands, show them now,
    # otherwise just show the formatted the difference arrays.
    s = (CRDB::Resource).format_diff_arrays(in_zimbra_only, in_database_only, in_both)

    if (sync_commands)
      changed_resources = []
      lines = s.split("\n")
      lines.each do |line|
        if (line =~ /^\s+(cal_.+)\*\s*$/)
          changed_resources.push($1)
        end
      end

      if (changed_resources.empty?)
        return ''
      else
        command_string = ''
        changed_resources.each do |changed_resource|
          command_string += "remctl dsi-crdb.stanford.edu crdb sync " +
                            "#{changed_resource} #{zimbra_server}" + "\n"
        end
        return command_string
      end
    else
      return s
    end

    return s
  end

  ### The table command
  def table()
    platform = platform()
    connect_info = (CRDB::Session).connect(platform)

    # Get the tablename
    if (ARGV.length >= 1)
      tablename = ARGV[1]
    else
      tablename = nil
    end

    tables_that_can_be_displayed =
    {
      'buildings'   => ['Building',   ['QB', 'short_name']],
      'building'    => ['Building',   ['QB', 'short_name']],
      'types'       => ['Type',       ['code', 'category_code']],
      'type'        => ['Type',       ['code', 'category_code']],
      'categories'  => ['Category',   ['code']],
      'category'    => ['Category',   ['code']],
      'policies'    => ['Policy',     ['name']],
      'policy'      => ['Policy',     ['name']],
      'departments' => ['Department', ['code', 'long_name']],
      'department'  => ['Department', ['code', 'long_name']],
      'adminunits'  => ['Adminunit',  ['code', 'long_name']],
      'adminunit'   => ['Adminunit',  ['code', 'long_name']],
    }

    return_value = ''

    # If tablename is nil return the list of displayable tables.
    if (tablename.nil?)
      return_value += "tables whose contents can be displayed:\n"
      tables_that_can_be_displayed.each do |key, value|
        return_value += '  ' + key + "\n"
      end

      return return_value
    else
      # tablename is not nil, so let's dump it.
      if (!tables_that_can_be_displayed.key?(tablename))
        raise "cannot display information for table '#{tablename}'"
      else
        class_name, columns_to_display = tables_that_can_be_displayed[tablename]

        # Prepend 'CRDB' to the class_name
        class_name = 'CRDB::' + class_name

        # Get all the rows of this table.
        rows = eval("#{class_name}.all")

        rows.each do |row|
          columns_to_display.each do |column|
            return_value += row.send(column) + ' '
          end
          return_value += "\n"
        end
      end

      return return_value
    end

  end


  ### The set command
  def set()
    platform = platform()
    connect_info = (CRDB::Session).connect(platform)

    # Get the fieldname and value
    uid       = nil
    fieldname = nil
    value     = nil

    if (ARGV.length >= 4)
      uid       = ARGV[1]
      fieldname = ARGV[2]
      value     = ARGV[3]
    elsif (ARGV.length == 1)
      return (CRDB::Misc).set_resource_attribute(uid, fieldname, value)
    else
      raise "you must supply a uid, fieldname, and value"
    end

    return (CRDB::Misc).set_resource_attribute(uid, fieldname, value)
  end

  ### The create command
  def create()
    options_hash = process_options()
    platform = options_hash['platform']
    dry_run  = options_hash['dry_run']

    connect_info = (CRDB::Session).connect(platform)

    return_value = ''

    # Get the full-path to the XML file.
    if (ARGV.length >= 2)
      xml_file  = ARGV[1]
    else
      raise "you must supply the full path to an XML file"
    end

    # Step 1. Create a temporary file.
    xml_file_temp      = Tempfile.new('tempxml')
    xml_file_temp_path = xml_file_temp.path()

    # Step 2. Create a Kerberos ticket file
    kerb = (CRDB::CRDBRemctl).make_kerberos()

    # Step 3. Copy the xml file to this temporary file
    command = "/bin/cp #{xml_file} #{xml_file_temp_path}"
    stdout, stderr = kerb.run(command)

    if (stderr.length > 0)
      raise "error copying #{xml_file} to #{xml_file_temp_path}: #{stderr}"
    end

    # Step 4. Create the resource
    xml_string = (CRDB::Misc).slurp(xml_file_temp_path)
    File.unlink(xml_file_temp_path)
    resource = (CRDB::Resource).from_xml(xml_string)

    # Step 5. Add a resourceadmin from contact_email
    contact_email = resource.contact_email
    kerb_uid = nil
    if (contact_email =~ /^([^@]+)@/)
      sunetid = $1

      kerb_uid = nil
      begin
        kerb_uid = (CRDB::Misc).sunetid_to_uid(sunetid,
                                     {:kerberos => kerb})
      rescue
        puts "There was error getting the kerberos id: " + $!
        kerb_uid = nil
      end

      if (!kerb_uid.nil?)
        resourceadmin = (CRDB::Resourceadmin).new()
        resourceadmin.sunetid = kerb_uid
        resource.resourceadmins << resourceadmin
      end
    end

    if (dry_run)
      return_value += resource.to_s
    else
      # Remove the resourceadmin before saving.
      resource.resourceadmins.clear

      # Now save the resource
      resource.save!

      # Now add the resourceadmins back.
      if (!kerb_uid.nil?)
        resourceadmin = (CRDB::Resourceadmin).new()
        resourceadmin.sunetid = kerb_uid
        resource.resourceadmins << resourceadmin
        resource.save!
      end

      return_value += "Resource saved!\n"
      return_value += resource.to_s
    end

    return return_value

  end

  def (CRDB::CRDBRemctl).make_kerberos()
    require 'yaml'

    # 1. Get the stage.
    stage = (CRDB::Misc).stage()

    # 2. Get the kerberos principal id from the stage.
    kerberos_yml = YAML::load_file('/etc/crdb/kerberos.yml')
    stage_to_kerberos_principal_id = kerberos_yml['stage_to_kerberos_principal_id']
    kerberos_principal_id = stage_to_kerberos_principal_id[stage]

    # 3. Get the kerberos config information
    kerb_config = kerberos_yml[kerberos_principal_id]

    # 4. Using the information in kerb_config, make the CRDB::Kerberos object.
    require 'crdb/kerberos'
    kerb = (CRDB::Kerberos).new()

    kerb.principal_realm    = kerb_config['realm']
    kerb.principal_primary  = kerb_config['primary']
    kerb.principal_instance = kerb_config['instance']
    kerb.keytab_file        = kerb_config['keytab_file']

    # Final. Return the CRDB::Kerberos object.
    return kerb
  end

  ### The system_check command
  def system_check()

    # Load the config, database, and kerberos YAML files
    config   = YAML::load_file('/etc/crdb/config.yml')
    database = YAML::load_file('/etc/crdb/database.yml')
    kerberos = YAML::load_file('/etc/crdb/kerberos.yml')

    # config
    stage   = config['stage']

    # version
    version = CRDB::Misc.get_crdb_version()

    # kerberos
    stage_to_kerberos_principal_id = kerberos['stage_to_kerberos_principal_id']
    kerberos_principal_id = stage_to_kerberos_principal_id[stage]
    kerb_config = kerberos[kerberos_principal_id]
    principal_realm    = kerb_config['realm']
    principal_primary  = kerb_config['primary']
    principal_instance = kerb_config['instance']
    keytab_file        = kerb_config['keytab_file']

    # Does the keytab file exist?
    keytab_file_exists = '[file does NOT exist!]'
    if (File.exist?(keytab_file))
      keytab_file_exists = '[file exists]'
    end

    # database
    development_config = database['development']
    production_config = database['production']




    results = <<EOR
GENERAL
version: #{version}
stage:   #{stage}

KERBEROS
kerb principal: #{kerberos_principal_id}
realm:          #{kerb_config['realm']}
primary:        #{kerb_config['primary']}
instance:       #{kerb_config['instance']}
keytab file:    #{kerb_config['keytab_file']} #{keytab_file_exists}

DATABASE
[development]
adapter:   #{development_config['adapter']}
host:      #{development_config['host']}
database:  #{development_config['database']}
username:  #{development_config['username']}
password:  [REDACTED]

[production]
adapter:   #{production_config['adapter']}
host:      #{production_config['host']}
database:  #{production_config['database']}
username:  #{production_config['username']}
password:  [REDACTED]
EOR

  end

 end # matches "class CRDBRemctl"
end # matches "module CRDB"
