require 'yaml'

module CRDB
 module Misc

  # Returns true if x is nil or empty.
  def (CRDB::Misc).nilorempty?(x)
    if (x.nil? || x.empty?)
      return true
    else
      return false
    end
  end

  # Returns nil if x is either nil or the empty string, x otherwise.
  def (CRDB::Misc).nillify(x)
    if (x.nil? || x.empty?)
      return nil
    else
      return x
    end
  end

  # This one is almost the same as nillify except it returns nil if x is
  # nil, the empty string, or all whitespace.
  def (CRDB::Misc).ws_nillify(x)
    if (x.nil? || x.strip.empty?)
      return nil
    else
      return x.strip
    end
  end

  # Return nil if nil or empty, otherwise return stripped.
  def (CRDB::Misc).strip_or_nillify(x)
    if (x.nil? || x.empty?)
      return nil
    else
      return x.strip
    end
  end


  # This function returns a String.
  # 1. If x is nil it returns the string 'nil'.
  # 2. If x is the empty string it returns the string 'empty'.
  # 3. If x is a non-empty string, returns the string surrounded by single
  #    quotes.
  def (CRDB::Misc).what_is_this_string(x)
    if (x.nil?)
      return 'nil'
    elsif (x.to_s.empty?)
      return 'empty_string'
    else
      return %Q['] + x.to_s + %Q[']
    end
  end


#  ## NO LONGER NEEDED
#  # Create a hash mapping "in production" emails to e-mails in the
#  # resources table
#  def (CRDB::Misc).in_production_to_in_resources
#    filehandle = File.open("../in_resource_to_in_production.txt", "r")
#
#    in_production_to_in_resource = {}
#    while (line = filehandle.gets)
#
#      # Skip any line starting with '#'
#      if (line =~ /^\s*\#/)
#        next
#      end
#
#      fields = line.split(/,;,/)
#
#      resource_email = fields[0]
#      production_email = fields[1]
#
#      in_production_to_in_resource[production_email] = resource_email
#    end
#
#    filehandle.close
#
#    return in_production_to_in_resource
#  end

  # Change some of the zimbra uid's in the resource tables by
  # giving them the uid used in production.
  def (CRDB::Misc).update_to_production_uids

    in_production_to_in_resource = (CRDB::Misc).in_production_to_in_resources()
    require 'crdb/resource'

    in_production_to_in_resource.each do |production_uid, resource_uid|

      # Get the database resource with this email address.
      r = (CRDB::Resource).find(:first, :conditions => {:uid => resource_uid})

      if (r.nil?)
        $stderr.puts("could not find resource in RESOURCES with uid '#{resource_uid}'")
        next
      end

      r.uid = production_uid

      r.save!
      puts "processed #{resource_uid} --> #{production_uid}"
    end

  end

  # Run a command and return an array consisting of
  # stdout and stderr.
  def (CRDB::Misc).run_command(command)
    require 'open3'

    # Turn off VERBOSE for a while
    save_verbose = $VERBOSE
    $VERBOSE = nil
    stdin, stdout, stderr = Open3.popen3(command)
    $VERBOSE = save_verbose

    # Dump output
    stdout_string = stdout.read
    stderr_string = stderr.read

    return [stdout_string, stderr_string]
  end


  # Given the argument to the -p (platform) option, returns
  # either 'production' or 'development'. If the argument matches
  # neither, raises an error.
  def (CRDB::Misc).arg_process_platform_option(poption_arg)

    platform = nil

    if (poption_arg =~ /^dev/i)
      platform = 'development'
    elsif (poption_arg =~ /^prod/i)
      platform = 'production'
    else
      raise "unrecognized platform '#{poption_arg}'"
    end

    return platform
  end

  # Takes a an options string from the command line
  # Valid options are
  #   '--help'
  def (CRDB::Misc).arg_process_common_options(opts)

    verbose     = false
    show_help   = false
    platform    = 'production'
    dry_run     = false
    list_fields = false

    opts.each do |opt, arg|
      case opt
        when '--help'
          show_help = true
        when '--platform'
          platform = arg
        when '--dry_run'
          dry_run = true
        when '--list_fields'
          list_fields = true
        when '--verbose'
          verbose = true
        else
          raise "unrecognized option '#{opt}'"
      end
    end

    begin
      platform =  (CRDB::Misc).arg_process_platform_option(platform)
    rescue
      puts "unrecognized platform '#{platform}' (should be either " +
           "'production' or 'development')"
      Process.exit
    end


    return {
            'help'        => show_help,
            'dry_run'     => dry_run,
            'platform'    => platform,
            'list_fields' => list_fields,
            'verbose'     => verbose
           }
  end

  # Returns true if the supplied string is a valid sunetid. Raises an
  # exception if no sunetid is provided.
  # Takes the :kerberos option.
  def (CRDB::Misc).sunetid?(sunetid, *options_array)
    if (sunetid.nil?)
      raise "missing argument"
    end

    if (options_array.nil?)
      options_array = {}
    end

    begin
      uid = (CRDB::Misc).sunetid_to_uid(sunetid, *options_array)
    rescue NoUidFoundException => exception
      # No uid found, so set uid to nil
      uid = nil
    rescue
      # All other errors
      raise
    end

    if (uid.nil?)
      return false
    else
      return true
    end

  end

  # In the next function we want to raise an exception when we attempt to
  # map a sunetid to an LDAP uid but no uid is found. We define a new
  # exception called CRDB::NoUidFoundException to raise in this situation.
  class CRDB::NoUidFoundException < RuntimeError
  end

  # Given a sunetid returns the uid associated with it
  # If it cannot find the uid, it raises the CRDB::NoUidFoundException
  # exception. If no sunetid is passed raises the default exception.
  #
  # Takes an the option :kerberos which should be a CRDB::Kerberos
  # object. If this option is omitted the ldapsearch command is run in the
  # current process context without any external kerberos privileges.
  def (CRDB::Misc).sunetid_to_uid(sunetid, *options_array)

    # Extract any options
    kerb = nil
    if (options_array.length > 0)
      options_hash = options_array[0]
      if (options_hash.key?(:kerberos))
        kerberos = options_hash[:kerberos]
      end
    end

    ldap_cmd = 'ldapsearch'
    cmd = %Q[#{ldap_cmd} -LLL -Q -h ldap -b "dc=stanford,dc=edu" suSunetID=#{sunetid}]
    # puts cmd

    if (kerberos)
      stdout, stderr = kerberos.run(cmd)
    else
      stdout, stderr = (CRDB::Misc).run_command(cmd)
    end


    if (!stderr.nil? && (stderr.length > 0))
      raise "error running ldap command '#{cmd}': " + stderr
    elsif (!stdout.nil? && (stdout.length > 10))
      # There appears to be no error in the call. Look for the uid.
      lines = stdout.split("\n")
      lines.each do |line|
        if (line =~ /^uid\:\s+(\S+)\s*$/)
          return $1
        end
      end

      # If we get here we did not find it.
       raise (CRDB::NoUidFoundException).new(),
             "did not find uid from ldap command '#{cmd}': [#{stdout}]"
    else
      return nil
    end


  end


  ### The set command
  def (CRDB::Misc).set_resource_attribute(uid = nil, fieldname = nil, value = nil)
    settable_fields =
    {
      'contact_fullname'    => 'single',
      'contact_email'       => 'single',
      'contact_phone'       => 'single',
      'building_short_name' => 'single',
      'capacity'            => 'single',
      'displayname_override' => 'single',
      'floor-room'          => 'single',
      'resourcephone'       => 'multiple',
      'nickname'            => 'single',
      'notes'               => 'single',
      'room_override'       => 'single',
      'resource_name'       => 'single',
      'old_sundial_name'    => 'single',
      'policy'              => 'single',
      'department'          => 'multiple',
      'adminunit'           => 'multiple',
      'resourceadmin'       => 'multiple',
      'type'                => 'single'
    }

    return_value = ''

    # If both fieldname and value are nil return the list of
    # fields that can be set.
    if (uid.nil? && fieldname.nil? && value.nil?)
      return_value += "the fields that can be set are\n"

      keys = settable_fields.keys.sort
      keys.each do |key|
        # Pad the spaces
        pad_length = 24 - key.length
        spacer = ' ' * pad_length
        return_value += "  #{key}#{spacer}(#{settable_fields[key]})\n"
      end

      return return_value
    elsif (uid.nil? || fieldname.nil? || value.nil?)
      raise "all three of uid, fieldname, and value must be non-empty"
    end

    # Is fieldname one that we can set? If not, raise an exception.
    if (!settable_fields.key?(fieldname))
      raise "fieldname '#{fieldname}' is not one of the fields I can set"
    end

    # Change some of the fieldnames
    external_to_internal_fieldname =
    {
      'floor-room' => 'FR'
    }

    external_fieldname = fieldname
    if (external_to_internal_fieldname.key?(fieldname))
      fieldname = external_to_internal_fieldname[fieldname]
    end

    # If we get here we have the three arguments. Is value equal to
    # "null"?
    value_is_null = false
    if (value =~ /^null$/i)
      value_is_null = true
      value = nil
    end

    # Get the resource (or, at least, try to).
    resource = (CRDB::Resource).find_by_uid(uid)
    if (resource.nil?)
      raise "could not find a resource with uid '#{uid}'"
    end

    # There are two cases: the field is multivalued and the value is
    # single-valued.
    if (settable_fields[external_fieldname] =~ /single/)
      # SINGLE-VALUED

      old_value = resource.send(fieldname)

      # We have to handle some of the fields in a custom manner.
      if (fieldname =~ /building.short.name/)
        # BUILDING
        building = (CRDB::Building).find_by_short_name(value)
        if (building.nil?)
          raise "could not find a building with short_name '#{value}'"
        else
          resource.building = building
          resource.save!
          return_value += "changed resource #{uid}'s building from '#{old_value.to_s}' " +
                          "to '#{resource.building_short_name}'"
        end
      elsif (fieldname =~ /policy/)
        # (SCHEDULING) POLICY
        policy = (CRDB::Policy).find_by_name(value)
        if (policy.nil?)
          raise "could not find a policy with name '#{value}'"
        else
          resource.policy = policy
          resource.save!
          return_value += "changed resource #{uid}'s policy from '#{old_value.name}' " +
                          "to '#{resource.policy.name}'"
        end
      elsif (fieldname =~ /type/)
        # TYPE
        type = (CRDB::Type).find_by_code(value)
        if (type.nil?)
          raise "could not find a type with code '#{value}'"
        else
          resource.type = type
          resource.save!
          return_value += "changed resource #{uid}'s type from '#{old_value.code}' " +
                          "to '#{resource.type.code}'"
        end
      else
        # EVERY OTHER SINGLE-VALUED FIELD

        # 25-Aug-2010: Add code to ensure that old_sundial_name can only be
        # changed to nil. (1.0.17)
        if ((fieldname =~ /old_sundial_name/) && (!value.nil?))
          raise "cannot set old_sundial_name to a non-null value"
        end

        resource.send("#{fieldname}=", value)
        resource.save!
        return_value += "changed resource #{uid}'s #{external_fieldname} from '#{old_value} " +
                        "to '#{value}'"
      end

    elsif (settable_fields[fieldname] =~ /multi/)
      # MULTIPLE-VALUED
      action = nil

      # Get the prefix to the value (should be either '+' or '-')
      if (value =~ /^(\+|\-)(.*)$/)
        action = $1
        value = $2
        if (action =~ /\+/)
          action = 'add'
        elsif (action =~ /\-/)
          action = 'remove'
        else
          raise "Huh?!?"
        end
      elsif (value_is_null)
        action = 'null'
      else
        raise "a multi-valued field must start with a '+' or '-'"
      end

      # NOT YET COMPLETE!
      if (action =~ /add/)
        # ADDING
        if (fieldname =~ /department/)
          ### add department
          department = (CRDB::Department).find_by_code(value)
          if (department.nil?)
            raise "could not find an department corresponding to '#{value}'"
          else
            # Add this department.
            resource.departments.delete(department)
            resource.departments << department
            resource.save!
            return_value += "added department #{department.code} to #{resource.uid}"
          end
        elsif (fieldname =~ /adminunit/)
          ### add adminunit
          adminunit = (CRDB::Adminunit).find_by_code(value)
          if (adminunit.nil?)
            raise "could not find an adminunit corresponding to '#{value}'"
          else
            # Add this adminunit.
            resource.adminunits.delete(adminunit)
            resource.adminunits << adminunit
            resource.save!
            return_value += "added adminunit #{adminunit.code} to #{resource.uid}"
          end
        elsif (fieldname =~ /resourceadmin/)
          ### add resourceadmin
          if (resource.resourceadmins.exists?(:sunetid => value))
            return return_value
          end

          resourceadmin = (CRDB::Resourceadmin).new()
          resourceadmin.sunetid  = value

          # Add this resourceadmin.
          resource.resourceadmins << resourceadmin
          resource.save!
          return_value += "added resourceadmin #{resourceadmin.sunetid} to #{resource.uid}"
        elsif (fieldname =~ /resourcephone/)
          ### add resourcephone
          if (resource.resourcephones.exists?(:phone => value))
            return return_value
          end

          resourcephone = (CRDB::Resourcephone).new()
          resourcephone.phone  = value

          # Add this resourcephone.
          resource.resourcephones << resourcephone
          resource.save!
          return_value += "added resourcephone #{resourcephone.phone} to #{resource.uid}"
        end
      elsif (action =~ /remove/)
        # REMOVING
        if (fieldname =~ /department/)
          ### remove department
          department = (CRDB::Department).find_by_code(value)
          if (department.nil?)
            raise "could not find a department corresponding to '#{value}'"
          else
            # Remove this department.
            resource.departments.delete(department)
            resource.save!
            return_value += "removed department #{department.code} from #{resource.uid}"
          end
        elsif (fieldname =~ /adminunit/)
          ### remove adminunit
          adminunit = (CRDB::Adminunit).find_by_code(value)
          if (adminunit.nil?)
            raise "could not find a adminunit corresponding to '#{value}'"
          else
            # Remove this adminunit.
            resource.adminunits.delete(adminunit)
            resource.save!
            return_value += "removed adminunit #{adminunit.code} from #{resource.uid}"
          end
        elsif (fieldname =~ /resourceadmin/)
          ### remove resourceadmin
          resourceadmin = resource.resourceadmins.find(:first,
                          :conditions => {:sunetid => value, :workgroup => nil})
          if (!resourceadmin.nil?)
            resource.resourceadmins.delete(resourceadmin)
            return_value += "removed resourceadmin# #{resourceadmin.sunetid} from #{resource.uid}"
          end
        elsif (fieldname =~ /resourcephone/)
          ### remove resourcephone
          resourcephone = resource.resourcephones.find(:first,
                          :conditions => {:phone => value})
          if (!resourcephone.nil?)
            resource.resourcephones.delete(resourcephone)
            return_value += "removed resourcephone# #{resourcephone.phone} from #{resource.uid}"
          end
        end
      elsif (action =~ /null/)
        # REMOVING ALL
      end
    end

    return return_value
  end

  def (CRDB::Misc).slurp(filename)
    f = File.open(filename, "r")

    data = ''
    f.each_line do |line|
      data += line
    end

    f.close
    return data
  end

  # What server stage are we running? Returns a string. Should be one of
  # 'dev', 'uat', or 'prod'.
  def (CRDB::Misc).stage()
   config = YAML::load_file('/etc/crdb/config.yml')
   return config['stage']
  end

  # Given an integral number of seconds, returns an array of integers
  # of the form [a (years), b (months), c (days), d (hours), e (minutes), f (seconds)]
  def (CRDB::Misc).seconds_to_time(seconds)
    seconds = seconds.to_i

    seconds_in_a_minute = 60
    seconds_in_an_hour  = 3600
    seconds_in_a_day    = 86400
    seconds_in_a_week   = 604800
    seconds_in_a_month  = 2592000
    seconds_in_a_year   = 31557600

    years = seconds / seconds_in_a_year
    seconds = seconds - (years * seconds_in_a_year)

    months = seconds / seconds_in_a_month
    seconds = seconds - (months * seconds_in_a_month)

    weeks = seconds / seconds_in_a_week
    seconds = seconds - (weeks * seconds_in_a_week)

    days = seconds / seconds_in_a_day
    seconds = seconds - (days * seconds_in_a_day)

    hours = seconds / seconds_in_an_hour
    seconds = seconds - (hours * seconds_in_an_hour)

    minutes = seconds / seconds_in_a_minute
    seconds = seconds - (minutes * seconds_in_a_minute)

    time = [years, months, weeks, days, hours, minutes, seconds]

    return time
  end


  # Format an array produced by (CRDB::Misc).seconds_to_time()
  # We only go back two magnitudes (i.e., if it has been a year we show
  # years and months, if it has been an hour we show hours ans minutes, etc.)
  def (CRDB::Misc).format_seconds_to_time_array(seconds_to_time_array)

    time_strings_array = []

    ## ###      ## ###      ## ###      ## ###      ## ###      ## ###      ## ###
    format_time_unit = Proc.new do |value, string|
      plural_suffix = (CRDB::Misc).sing_or_plural(value)
      rv = value.to_s + ' ' + string + plural_suffix
    end
    ## ###      ## ###      ## ###      ## ###      ## ###      ## ###      ## ###

    if    (seconds_to_time_array[0] > 0)
      time_strings_array.push(format_time_unit.call(seconds_to_time_array[0], 'year'))
      time_strings_array.push(format_time_unit.call(seconds_to_time_array[1], 'month'))
    elsif (seconds_to_time_array[1] > 0)
      time_strings_array.push(format_time_unit.call(seconds_to_time_array[1], 'month'))
      time_strings_array.push(format_time_unit.call(seconds_to_time_array[2], 'week'))
    elsif (seconds_to_time_array[2] > 0)
      time_strings_array.push(format_time_unit.call(seconds_to_time_array[2], 'week'))
      time_strings_array.push(format_time_unit.call(seconds_to_time_array[3], 'day'))
    elsif (seconds_to_time_array[3] > 0)
      time_strings_array.push(format_time_unit.call(seconds_to_time_array[3], 'day'))
      time_strings_array.push(format_time_unit.call(seconds_to_time_array[4], 'hour'))
    elsif (seconds_to_time_array[4] > 0)
      time_strings_array.push(format_time_unit.call(seconds_to_time_array[4], 'hour'))
      time_strings_array.push(format_time_unit.call(seconds_to_time_array[5], 'minute'))
    elsif (seconds_to_time_array[5] > 0)
      time_strings_array.push(format_time_unit.call(seconds_to_time_array[5], 'minute'))
      time_strings_array.push(format_time_unit.call(seconds_to_time_array[6], 'second'))
    elsif (seconds_to_time_array[6] >= 0)
      time_strings_array.push(format_time_unit.call(seconds_to_time_array[6], 'second'))
    end

    return_value = time_strings_array.join(", ")

    return return_value
  end


  def (CRDB::Misc).seconds_to_time_string(seconds)
    time_seconds_array = (CRDB::Misc).seconds_to_time(seconds)
    return (CRDB::Misc).format_seconds_to_time_array(time_seconds_array)
  end

  # Returns 's' is the argument is 0 or larger than 1, '' empty otherwise.
  def (CRDB::Misc).sing_or_plural(n)
    m = n.to_i
    if (m == 1)
      return ''
    else
      return 's'
    end
  end

  # Returns the current version of the CRDB software
  def (CRDB::Misc).get_crdb_version
    require 'crdb'
    return CRDB::VERSION
  end

  # Returns the stage as defined in the config.yml file.
  def (CRDB::Misc).get_crdb_stage
    config = YAML::load_file('/etc/crdb/config.yml')
    stage  = config['stage']
    return stage
  end

 end # matches "module Misc"
end # matches "module CRDB"
