# History:
#
# 17-Aug-2009: Added a new validation displayname_is_valid to ensure that
# no weird characters get in the display name (this could cause problems
# for Zimbra). See also Jira #CRDB-7
#
# 26-Feb-2010: Added a 'before_save' callback to convert several of the
# attributes to nil if they are empty or all whitespace. (1.0.15)
#
# 11-Mar-2010: Fixed bug when searching for department. Changed name of
# method "adminunit" to "adminunits_formatted". (1.0.16)
#
# 25-Aug-2010: Removed the readonly restriction on
# old_sundial_name. Instead, in def (CRDB::Misc).set_resource_attribute
# we only allow setting old_sundial_name to nil. (1.0.17)
#
# 01-Oct-2010: Added to the uid format validity check that the length must
# be no greater than 60 characters. (1.0.18)

module CRDB
 class Resource < ActiveRecord::Base

  # Set the "old_sundial_name" field to be readonly. This means that the
  # field can be set upon creation but not on update.
  # 25-Aug-2010: Remove the readonly restriction on
  # old_sundial_name. Instead, in def (CRDB::Misc).set_resource_attribute
  # only allow setting to nil.
  # attr_readonly :old_sundial_name

  # In case this is running on a server where the active record or active
  # support libraries are in /usr/lib/ruby/1.8/:
  # $:.push('/usr/lib/ruby/1.8/')

  require 'crdb/building'
  require 'crdb/policy'
  require 'crdb/type'
  require 'crdb/category'
  require 'crdb/department'
  require 'crdb/ownership'
  require 'crdb/adminunit'
  require 'crdb/adminownership'
  require 'crdb/resourceadmin'
  require 'crdb/event'
  require 'crdb/resourcephone'

  require 'crdb/misc'

  # Here are the foreign keys:
  belongs_to :type, :class_name => "CRDB::Type"
  belongs_to :policy, :class_name => "CRDB::Policy"
  belongs_to :building, :class_name => "CRDB::Building"

  # The "join table" ownerships links resources to departments
  has_many :ownerships, :class_name => "CRDB::Ownership", :dependent => :destroy
  has_many :departments, :class_name => "CRDB::Department", :through => :ownerships

  # The "join table" adminownerships links resources to adminunits
  has_many :adminownerships, :class_name => "CRDB::Adminownership", :dependent => :destroy
  has_many :adminunits, :class_name => "CRDB::Adminunit", :through => :adminownerships

  # CRDB::Resources (can) have many admins.
  has_many :resourceadmins, :class_name => "CRDB::Resourceadmin", :dependent => :destroy

  # CRDB::Resources (can) have many resourcephones.
  has_many :resourcephones, :class_name => "CRDB::Resourcephone", :dependent => :destroy

  # CRDB::Resources have many events
  has_many :events, :class_name => "CRDB::Event", :dependent => :destroy
  # Note. The ":dependent => :destroy" option deletes any events associated with
  # this resource when the resource is deleted.



  ## VALIDATIONS
  validates_uniqueness_of :uid, :scope => :domain
  validates_format_of :uid, :with => /\A[a-z0-9_]{1,60}\z/, :allow_nil => false,
    :message => "must be no longer than 60 characters and contain " +
                "only letters, numbers, and the underscore character"
  validates_format_of :FR, :with => /\A[A-Z0-9]+-[A-Z0-9]+\z/, :allow_nil => true
  validates_numericality_of :capacity, :allow_nil => true

  # At least one of contact_fullname, contact_email, or contact_phone must
  # be nonnull.
  validate :must_have_contact_info
  def must_have_contact_info
    if (contact_fullname.nil? && contact_email.nil? && contact_phone.nil?)
      errors.add_to_base("At least one of contact_fullname, contact_email, " +
                         "or contact_phone must be nonnull")
    end
  end

  validate :contact_email_is_valid
  def contact_email_is_valid
    if (!(CRDB::Misc).nilorempty?(contact_email))
      if (contact_email !~ /\A([^@\s]+)@((?:[-a-z0-9]+\.)+[a-z]{2,})\z/i)
        msg = "contact_email #{contact_email} looks to be invalid"
        errors.add_to_base("msg")
      end
    end
  end

  validate :contact_phone_is_valid
  def contact_phone_is_valid
    if (!(CRDB::Misc).nilorempty?(contact_phone))
      if (contact_phone !~ /\A\d\d\d\-\d\d\d\-\d\d\d\d\z/)
        msg = "contact_phone #{contact_phone} looks to be invalid"
        errors.add_to_base(msg)
      end
    end
  end

  validate :uid_is_valid
  def uid_is_valid
    # uid should contain only lowercase letters, numbers, and the
    # underscore character.
    if (uid !~ /^[a-z0-9_]+/)
      msg = "uid #{uid} looks to be invalid"
      errors.add_to_base(msg)
    end
  end

  validate :displayname_is_valid
  def displayname_is_valid
    if (display_name !~ /\A[a-z0-9[:blank:]\_\-\.\#\&\(\)]+\z/i)
      msg = "display name '#{display_name}' contains some bad characters" +
            "(only letters, numbers, spaces, and the symbols " +
            "_ (underscore), - (hyphen), . (period), # (hash mark), & (ampersand), " +
            "( ) (left and right parentheses) are allowed"
      errors.add_to_base(msg)
    end
  end

  validate :resource_phone_location_only
  def resource_phone_location_only
    # Get all the resourcephones.
    resourcephones = self.resourcephones
    if (!resourcephones.nil? && (resourcephones.length > 0))
      # There are resource phones. Is this a location resource?
      if (!self.location?)
        msg = "you cannot add resource phones to a non-location resource"
        errors.add_to_base(msg)
      end
    end
  end


  # Some extra Zimbra attributes ("za" stands for "Z"imbra "a"ttributes)
  attr_accessor :za_zimbraCalResAutoAcceptDecline
  attr_accessor :za_zimbraCalResAutoDeclineIfBusy
  attr_accessor :za_zimbraCalResAutoDeclineIfBusy
  attr_accessor :za_zimbraCalResAutoDeclineRecurring
  attr_accessor :za_zimbraCalResSite
  attr_accessor :za_zimbraCalResBuilding
  attr_accessor :za_zimbraCalResFloor
  attr_accessor :za_zimbraCalResRoom
  attr_accessor :za_zimbraCalResLocationDisplayName
  attr_accessor :za_zimbraCalResType
  attr_accessor :za_displayName
  attr_accessor :za_description
  attr_accessor :za_zimbraNotes

  # Want an indicator that the resource is "fake".
  attr_accessor :is_fake

  ## CALLBACKS
  after_create :add_event_after_create
  after_update :add_event_after_update

  def to_s
    # Format the building information
    building = self.building
    if (!building.nil?)
      building_short_name = building.short_name
      room = self.room
    else
      building_short_name = 'N/A'
      room = 'N/A'
    end

    depts_formatted = self.departments_formatted

    admin_units = self.adminunits
    adminunits_formatted = (CRDB::Adminunit).format_multiple_adminunits(admin_units)

    resourceadmins = self.resourceadmins
    resourceadmins_formatted = (CRDB::Resourceadmin).format_multiple_resourceadmins(resourceadmins)

    resorucephones = self.resourcephones
    resourcephones_formatted = (CRDB::Resourcephone).format_multiple_resourcephones(resourcephones)

    if (self.type.nil?)
      type_formatted = "UNKNOWN TYPE"
    else
      type_formatted = "#{self.type.code} (#{self.type.description})"
    end

    if (self.policy.nil?)
      policy_formatted = "UNKNOWN POLICY"
    else
      policy_formatted = "#{self.policy.name}"
    end

    floor = self.floor
    if (floor.nil?)
      floor = 'N/A'
    end

    qbfr = self.qbfr()
    if (qbfr.nil?)
      qbfr = 'N/A'
    end

    last_sync_time = self.last_sync_time
    time_since = ''
    if (!last_sync_time.nil?)
      seconds_since = Time.now - last_sync_time
      time_since = '(' + (CRDB::Misc).seconds_to_time_string(seconds_since) + ')'
    else
      last_sync_time = 'N/A'
    end

    return <<-EOS
id:               #{self.id}
uid:              #{self.uid}
domain:           #{self.domain}
display name:     #{self.display_name}
building:         #{building_short_name}
Q-B-F-R:          #{qbfr}
room:             #{room}
floor:            #{floor}
contact name:     #{self.contact_fullname}
contact e-mail:   #{self.contact_email}
contact phone:    #{self.contact_phone}
capacity:         #{self.capacity}
resource phones:  #{resourcephones_formatted}
policy:           #{policy_formatted}
old Sundial name: #{self.old_sundial_name}
nickname:         #{self.nickname}
resource name:    #{self.resource_name}
departments:      #{depts_formatted}
admin orgs:       #{adminunits_formatted}
type:             #{type_formatted}
description:      #{self.description}
notes:            #{self.notes}
resource admins:  #{resourceadmins_formatted}
display name override:        #{self.displayname_override}
zimbra location display name: #{self.zimbraLocationDisplayName}
last sync time:   #{last_sync_time} #{time_since}
    EOS
  end

  def category
    type = self.type
    if (!type.nil?)
      return type.category
    else
      return nil
    end
  end

  def zimbraCalResType
    if (!@za_zimbraCalResType.nil?)
      return @za_zimbraCalResType
    else
      category = self.category
      if (category.nil?)
        return nil
      elsif (category.location?)
        return 'Location'
      else
        return 'Equipment'
      end
    end
  end

  def building_short_name
    building = self.building

    if (building.nil?)
      return nil
    else
      return building.short_name
    end
  end

  def policy_name
    policy = self.policy

    if (policy.nil?)
      return nil
    else
      return policy.name
    end
  end

  def resourceadmin
    resourceadmins = self.resourceadmins
    return (CRDB::Resourceadmin).format_multiple_resourceadmins(resourceadmins)
  end

  def building_very_short_name
    building = self.building

    if (building.nil?)
      return nil
    else
      return building.very_short_name
    end
  end

  def room()
    if (!(CRDB::Misc).nilorempty?(self.room_override))
      return self.room_override
    else
      fr = self.FR()

      if ((CRDB::Misc).nilorempty?(fr))
        return nil
      elsif (fr =~ /\A\w+\-(\w+)\z/)
        return $1
      else
        return nil
      end
    end
  end

  def floor()
    if (!(CRDB::Misc).nilorempty?(self.room_override))
      return nil
    else
      fr = self.FR()

      if ((CRDB::Misc).nilorempty?(fr))
        return nil
      elsif (fr =~ /\A(\w+)\-\w+\z/)
        return $1
      else
        return nil
      end
    end
  end

  # Returns 'QBFR'
  def qbfr()
    building = self.building
    if (!building.nil?)
      qb = building.QB
      fr = self.FR

      if (!fr.nil?)
        return qb + '-' + fr
      else
        return qb
      end
    end
     return nil
  end


  # To get the Zimbra "name", we concatenate the uid and the domain.
  def zimbra_name
    uid    = self.uid
    domain = self.domain
    return uid + '@' + domain
  end

  # Return the attribute zimbraCalResLocationDisplayName, if it is set,
  # otherwise construct it from the building and room.
  def zimbraLocationDisplayName
    if (!self.za_zimbraCalResLocationDisplayName.nil?)
      return za_zimbraCalResLocationDisplayName
    else
      parts = []

      if (!self.building.nil?)
        parts.push("Building: #{self.building_short_name}")
      end

      if (!self.room.nil?)
        parts.push("Room: #{self.room}")
      end

      if (parts.length > 0)
        return parts.join(", ")
      else
        return nil
      end
    end
  end

  def restricted?
    p = self.policy()
    if (p.nil?)
      return false
    elsif (p.name =~ /manual/)
      return true
    else
      return false
    end
  end

  # Return the formatted department(s)
  def departments_formatted
    depts = self.departments
    return (CRDB::Department).format_multiple_departments(depts)
  end

  # Return the formatted adminunit(s)
  def adminunits_formatted
    adminunits = self.adminunits
    return (CRDB::Adminunit).format_multiple_adminunits(adminunits)
  end

  # Returns the type code
  def type_code
    return self.type.code
  end

  # Create the uid from a supplied name string.
  # Returns a string such as 'cal_lo_polya_hall_152'
  def derive_uid_from_name(name)
    # Step 1. Get the category code.
    if (category.nil?)
      raise "missing category for resource (#{name})"
    end

    category = self.category
    category_code = category.code

    # Step 2. Sanitize the name
    sanitized_name = (CRDB::Resource).sanitize_for_uid(name)

    # Step 3. Put 'em together
    uid = 'cal_' + category_code + '_' + sanitized_name

    return uid
  end


  # Takes a string and converts into a form suitable for a Zimbra uid
  def (CRDB::Resource).sanitize_for_uid(string_to_sanitize)

    # 1. Multiple consecutive spaces converted to a single space
    # 2. Single spaces converted to underscores.
    # 3. Dashes are coverted to underscores.
    # 4. Anything that is not in the "acceptable" list of
    #    characters is deleted.
    # 5. Replace multiple underscores with a single underscore.
    # 6. Remove leading or trailing underscores.
    # 7. Convert to lower case.

    @name = string_to_sanitize.clone

    # 1. Remove duplicate spaces.
    @name.squeeze!(" ")

    # 2. Replace spaces with underscore
    @name.gsub!(/\s/, '_')

    # 3. Replace dashes with underscore
    @name.gsub!(/\-/, '_')

    # 4. Delete anything other than letter, number, and underscore.
    @name.gsub!(/[^a-zA-Z0-9\_]/, '')

    # 5. Replace multiple underscores with a single underscore
    @name.squeeze!("_")

    # 6. Delete trailing or leading underscores
    @name.gsub!(/\A\_/, '')
    @name.gsub!(/\_$/, '')

    # 7. Make lower case
    @name.downcase!()

    return @name
  end

  def location?
    category = self.category
    if (!category.nil?)
      # 1. First guess via the cateogory.
      category_code = category.code

      if (category_code =~ /lo/)
        return true
      else
        return false
      end
    else
      # 2. Now try the uid
      if (self.uid =~ /cal_lo/)
        return true
      else
        return false
      end
    end
  end

  def location_display_name
    building = self.building_short_name()
    room     = self.room()
    nickname = self.nickname()

    if ((CRDB::Misc).nilorempty?(building))
      building = 'NO BUILDING'
    end

    if (!(CRDB::Misc).nilorempty?(room))
      room = ' ' + room
    else
      room = ''
    end

    if (!(CRDB::Misc).nilorempty?(nickname))
      nickname = ' ' + nickname
    else
      nickname = ''
    end

    return building + room + nickname
  end


  def non_location_display_name

    # Get the owning departments (formatted into a single string)
    depts_formatted = self.departments_formatted

    resource_name  = self.resource_name()

    type = self.type()
    if (!type.nil?)
      type_code = type.code()
    else
      type_code = "UNKNOWN TYPE"
    end

    if (resource_name.nil? || resource_name.empty?)
      resource_name = 'NO_RESOURCE_NAME' ;
    end

    return "#{depts_formatted} #{type_code} #{resource_name}"
  end


  def display_name
    # If there is a displayname_override value, use that.
    if (!self.displayname_override.nil?)
      return self.displayname_override
    elsif (self.location?)
      return self.location_display_name()
    else
      return self.non_location_display_name()
    end
  end


#  # Return an array of CRDB::Resource objects, one for each
#  # resource in the specified Zimbra server
#  def get_all_zimbra_resource(zimbra_server)
#    if ((CRDB::Misc).nilorempty?(zimbra_server))
#      raise "cannot get all Zimbra resources without a server"
#    end
#
#    remctl = (CRDB::RemctlCaladmin).new()
#    remctl.server = zimbra_server
#
#    string = remctl.dump_attributes()
#
#  end


  # This is ONLY for migration and should be deleted afterwards
  def migration_room_override(qbfr)
    qbfr_to_room_override = {
      '14-705-01-500' => 'Conference Room',
      '90-345R-01-BRZWY' => 'Breezeway',
      '90-345R-01-CY' => 'Courtyard',
      '05-200-01-LNGE' => 'Lounge',
      '07-340-01-PAT' => 'Courtyard',
      '07-590-01-A' => 'Atrium',
      '07-590-01-L' => 'Lawn',
      '01-260-01-CY' => 'Courtyard',
      '07-306-01-PAT' => 'Courtyard',
      '07-570-01-PAT' => 'Foyer',
      '08-450-01-PAT' => 'Patio',
      '14-160-01-PAT' => 'Lawn',
      '07-570-SB1-CY' => 'Light Court',
      '07-307-01-PAT' => 'Courtyard Lawn',
      '03-420-00-001' => 'Patio',
      '04-070-01-PATIO' => 'Social Entry (Lobby)',
      '04-070-01-PLZA' => 'Courtyard',
      '04-070-02-TER' => 'Second Floor Terrace',
      '04-070-03-TER' => 'Third Floor Terrace',
      '01-420-01-LNGE' => 'Lounge',
      '01-420-01-PAT' => 'Patio',
      '06-010-05-LNGE' => 'Lounge',
      '05-510-02-AUD' => 'Auditorium'
    }

    if (qbfr_to_room_override.key?(qbfr))
      return qbfr_to_room_override[qbfr]
    else
      return nil
    end
  end

  # Derive the Zimbra description from the other attributes.
  # However, if za_description is set, use that instead.
  # Here is the recipe:
  # (see ??? doc for details).
  def description

    if (!self.za_description.nil?)
      return self.za_description
    end

    pieces = []

    # 1 and 2. Get all the adminunit information
    admin_codes = []
    admin_long_names = []
    adminunits = self.adminunits
    adminunits.each do |adminuit|
      admin_codes.push(adminuit.code)
      admin_long_names.push(adminuit.long_name)
    end

    # 3. CRDB::Department codes
    department_codes = []
    departments = self.departments
    departments.each do |department|
      department_codes.push(department.code)
    end

    # 4. CRDB::Resource type code
    type = self.type
    if (type.nil?)
      type_code = "UNKNOWN_TYPE"
    else
      type_code = type.code
    end

    # 5. QBFR
    qbfr = self.qbfr

    # 6. Old Sundial name
    old_sundial_name = self.old_sundial_name

    # 7. Any resource phones
    resourcephones = self.resourcephones
    phones = resourcephones.collect {|r| r.phone}


    # Put them together
    if (resourcephones.length > 0)
      pieces.push("ph: " + phones.join(", "))
    end

    if (admin_codes.length > 0)
      pieces.push(admin_codes.join(", "))
    end

    if (admin_long_names.length > 0)
      pieces.push(admin_long_names.join(", "))
    end

    if (department_codes.length > 0)
      pieces.push(department_codes.join(", "))
    end

    pieces.push(type_code)

    if (!qbfr.nil?)
      pieces.push(qbfr)
    end

    if (!old_sundial_name.nil?)
      pieces.push(old_sundial_name)
    end

    # Join with semicolons
    description = pieces.join("; ")

    return description
  end

  # Derive the scheduling policy from the zimCalRes attributes
  def derive_policy_from_attributes
    if ( @za_zimbraCalResAutoAcceptDecline &&
         @za_zimbraCalResAutoDeclineIfBusy &&
        !@za_zimbraCalResAutoDeclineRecurring)
      self.policy = (CRDB::Policy).find(:first, :conditions => {:name => 'auto_accept_unless_busy'})
    elsif ( @za_zimbraCalResAutoAcceptDecline &&
           !@za_zimbraCalResAutoDeclineIfBusy &&
           !@za_zimbraCalResAutoDeclineRecurring)
      self.policy = (CRDB::Policy).find(:first, :conditions => {:name => 'auto_accept_all'})
    elsif (!@za_zimbraCalResAutoAcceptDecline &&
           !@za_zimbraCalResAutoDeclineIfBusy &&
           !@za_zimbraCalResAutoDeclineRecurring)
      self.policy = (CRDB::Policy).find(:first, :conditions => {:name => 'manually_accept'})
    else
      raise "cannot derive policy!"
    end
  end


  # Let uid_to_resource1 be a hash mapping uid to a CRDB::Resource object, and
  # let email_to_resource2 another, similar hash.

  # Use:
  #   in_zimbra_only, in_database_only, in_both = (CRDB::Resource).find_differences()

  # Returns a report on the differences between the resources in Zimbra
  # and the resources in the database.

  # The report has this format:

  # Section 1: A hash of resources on the Zimbra server but not in the
  # database. The hash maps the Zimbra server full resource's uid
  # to its corresponding CRDB::Resource object.

  # Section 2: A hash of resources in the Database but not in the Zimbra server
  # The hash maps the database resource's uid to its corresponding
  # CRDB::Resource object.

  # Section 3. A hash of resources that are in the database AND in the
  # Zimbra server. The hash maps the resource's zimbra_name to the
  # pair (R_Z, R_D) where R_Z is the Zimbra CRDB::Resource object and R_D is the
  # database object.

  # Calls zimbra remctl commands so there must be valid kerberos
  # credentials available.

  # Options: :kerberos => kerberos_object.
  def (CRDB::Resource).find_differences(zimbra_server, *options_array)
    in_zimbra_only   = {}
    in_database_only = {}
    in_both          = {}

    # Extract any options and process
    caching = false
    kerberos = nil
    if (options_array.length > 0)
      options_hash = options_array[0]

      if (options_hash.key?(:caching))
        caching = options_hash[:caching]
      end

      if (options_hash.key?(:kerberos))
        kerberos = options_hash[:kerberos]
      end
    end

    # 1. Get all the resources in the database.
    database_resources = (CRDB::Resource).find(:all)
    resources_in_database = {}
    database_resources.each do |resource|
      resources_in_database[resource.zimbra_name] = resource
    end


    # 2. Get all the resources in Zimbra
    require 'crdb/zimbra'
    zimbra = (CRDB::Zimbra).new()
    zimbra.server = zimbra_server
    if (caching)
      zimbra.caching_on()
    end

    if (!kerberos.nil?)
      zimbra.kerberos = kerberos
    end

    resources_in_zimbra = zimbra.get_zimbra_name_to_resource()

    # Find the resources in Zimbra but not in the database.
    resource_zimbra_names_in_zimbra_only = resources_in_zimbra.keys() - resources_in_database.keys()
    resource_zimbra_names_in_zimbra_only.each do |zimbra_name|
      in_zimbra_only[zimbra_name] = resources_in_zimbra[zimbra_name]
    end

    # Find the resources in the database but not in Zimbra
    resource_zimbra_names_in_database_only = resources_in_database.keys() - resources_in_zimbra.keys()
    resource_zimbra_names_in_database_only.each do |database_zimbra_name|
      in_database_only[database_zimbra_name] = resources_in_database[database_zimbra_name]
    end

    # Find the resources in both.
    resource_zimbra_names_in_both = resources_in_database.keys() & resources_in_zimbra.keys()
    resource_zimbra_names_in_both.each do |zimbra_name|
      in_both[zimbra_name] =
                     [resources_in_zimbra[zimbra_name], resources_in_database[zimbra_name]]
    end

    return [in_zimbra_only, in_database_only, in_both]
  end

  # Format the output of (CRDB::Resource).find_differences suitable for printing.
  def (CRDB::Resource).format_diff_arrays(in_zimbra_only, in_database_only, in_both)

    rv = ''
    prefix = '  '

    rv += "The following resources are in Zimbra only:\n"
    in_zimbra_only.each do |uid, resource|
      rv += prefix + resource.uid + "\n"
    end

    rv += "The following resources are in the database only:\n"
    in_database_only.each do |uid, resource|
      rv += prefix + resource.uid + "\n"
    end

    rv += "The following resources are in Zimbra AND the database:\n"
    in_both.each do |uid, resource_pair|
      resource_zimbra   = resource_pair[0]
      resource_database = resource_pair[1]

      diffs = resource_zimbra.diff(resource_database)

      # No "\n" yet.
      rv += prefix + resource_database.uid

      # We only care about a few of the attributes.
      attributes = [
                     "contact_fullname", "contact_email",
                     "contact_phone", "policy",
                     "display_name", "capacity", "description",
                     "building_very_short_name", "room",
                     "zimbraLocationDisplayName",
                   ]
      changes_string = ''
      attributes.each do |attribute|
        if (diffs.key?(attribute))
          changes_string += prefix + prefix + "#{attribute} differs: "
          changes_string += "(zimbra, database) = "
          values = diffs[attribute]

          # Call method 'to_s_short' if the object has this method,
          # otherwise call 'to_s'.
          values_formatted = values.map do |value|
            if (value.nil?)
              ''
            else
              methods = value.methods()
              if (methods.include?('to_s_short'))
                value.to_s_short
              else
                value.to_s
              end
            end
          end

          changes_string += "('#{values_formatted[0]}', '#{values_formatted[1]}')\n"
        end
      end
      if (changes_string.length > 0)
        rv += "*"
      end
      rv += "\n" + changes_string
    end

    return rv
  end

  # Find the difference between two resources.
  # Use:
  #  resource1.diff(resource2)
  #
  # Returns a hash mapping resource attributes to a pair.
  #
  # Example:
  #    'resource_name' => ['Maryland', 'MaryLAnd'] a
  def diff(resource2)
    # Here are the attributes we want to examine:
    attributes_to_check = ['zimbra_name', 'domain', 'contact_fullname',
                           'contact_email', 'contact_phone',
                           'display_name','FR', 'room_override',
                           'capacity', 'nickname', 'resource_name',
                           'old_sundial_name', 'type', 'building',
                           'adminunits', 'departments', 'policy',
                           'description', 'building_very_short_name', 'room',
                           'zimbraLocationDisplayName'
                          ]
    differing_attribute_to_pair = {}
    attributes_to_check.each do |attribute|
      value1 = self.send(attribute)
      value2 = resource2.send(attribute)
      if (!value1.eql?(value2))
        #puts "#{attribute}: '#{value1.to_s}' is not the same as '#{value2.to_s}'"
        differing_attribute_to_pair[attribute] = [value1, value2]
      end
    end

    return differing_attribute_to_pair
  end


  # Take an XML version of a resource and convert to a resource object.
  def to_xml
    # TO DO.
  end


  # Given an XML string, extract the resource information and populate a
  # CRDB::Resource object.
  #
  # Returns a CRDB::Resource object.
  #
  # Note that the CRDB::Resource objects will not be SAVED!! You must do that yourself.
  #
  # Example:
  # resource = (CRDB::Resource).from_xml(xml_string)
  #
  # Options:
  #
  def (CRDB::Resource).from_xml(xml_string, *options_array)
    require 'rexml/document'
    include REXML

    # Extract any options and process
    error_checking = true
    if (options_array.length > 0)
      options_hash = options_array[0]

      if (options_hash.key?(:error_checking))
        error_checking = options_hash[:error_checking]
      end
    end

    doc = Document.new xml_string

    xml_resource_tag_name = 'Resource_Information'

    root = doc.root

    # If the root is "Resource_Information", then we are ready to
    # parse. If not, see if the root is calResourceRequest
    if (root.name == xml_resource_tag_name)
      elements = root.elements
      resource = (CRDB::Resource).new()

      uid           = nil
      category      = nil
      type1         = nil
      building      = nil
      adminunit     = nil
      department    = nil
      resourcephone = nil

      elements.each do |element|
        name  = element.name

        value = element.text
        value.strip!

        ### uid
        if (name == 'uid')
          # Usually this is blank, but if not, set it.
          if (!value.nil? && !value.empty?)
            uid = value.strip
          end
        ### QBFR
        elsif (name == 'QBFR')
          qb, fr = (CRDB::Building).parseQBFR(value)
          resource.FR = fr
          # Get Q
          q = nil
          if (qb =~ /\A(\d\d)/)
            q = $1
          end
          if (!q.nil?)
            building = (CRDB::Building).find_by_QB(qb)
            if (building.nil?)
              raise "could not find building corresponding to #{qb}"
            else
              # puts building.to_s
            end
          end
        ### contact information
        elsif (name == 'contact_information')
          # Get the elements
          contact_elements = element.elements
          contact_elements.each do |contact_element|
            contact_element_name  = contact_element.name
            contact_element_value = contact_element.text
            contact_element_value.strip!
            if (!contact_element_value.nil? && !contact_element_value.empty?)
              resource.send(contact_element_name + '=', contact_element_value)
            end
          end
        ### adminunit
        elsif (name == 'adminunit_code')
          adminunit = (CRDB::Adminunit).find_by_code(value)
          if (adminunit.nil?)
            raise "could not find adminunit corresponding to #{value}"
          end
        ### department
        elsif (name == 'department_code')
          if (!value.nil? && !value.empty?)
            department = (CRDB::Department).find_by_code(value)
            if (department.nil?)
              raise "could not find department corresponding to #{value}"
            end
          end
        ### category
        elsif (name == 'category_code')
          category = (CRDB::Category).find_by_code(value)
        ### type
        elsif (name == 'type_code')
          type1 = (CRDB::Type).find_by_code(value)
        ### resourcephone
        elsif (name == 'resourcephone')
          if (!value.nil? && !value.empty?)
            resourcephone = (CRDB::Resourcephone).new()
            resourcephone.phone = value
          end
        ### Everything else
        else
          value = (CRDB::Misc).strip_or_nillify(value)
          resource.send(name + '=', value)
        end
      end

      # old_sundial_name should be IGNORED
      old_sundial_name = nil

      # Add the type
      resource.type = type1

      # Add the building
      if (!building.nil?)
        resource.building = building
      end

      # Add the policy
      policy = nil
      case category.code
        when 'lo'
          policy = (CRDB::Policy).find_by_name('auto_accept_unless_busy')
        when 'eq'
          policy = (CRDB::Policy).find_by_name('auto_accept_unless_busy')
        when 'sc'
          policy = (CRDB::Policy).find_by_name('auto_accept_all')
        else
          raise "unrecognized category #{category.code}"
      end

      if (policy.nil?)
        raise "could not find policy for category #{category.code}"
      else
        resource.policy = policy
      end

      # Add the adminunit.
      resource.adminunits.push(adminunit)

      # Add the department (if it exists)
      if (!department.nil?)
        resource.departments.push(department)
      end

      # Add the resourcephone (if it exists)
      if (!resourcephone.nil?)
        resource.resourcephones.push(resourcephone)
      end

      # Add or derive the uid.
      if (!uid.nil?)
        resource.uid = uid
      else
        resource.uid = resource.derive_uid_from_name(resource.display_name)
      end


      # Do some error checking (but only if error_checking is true).
      if (error_checking)
        # Error check 1: If the resource is a non-location and the department
        # is empty, raise an error.
        if ((!resource.location?) && (resource.departments.nil? || resource.departments.empty?))
          raise "this is a non-location resource but is missing a department"
        end

        # Error check 2: If the resource is a non-location and the resource_name
        # is empty, raise an error.
        if ((!resource.location?) && (resource.resource_name.nil? || resource.resource_name.empty?))
          raise "this is a non-location resource but is missing a resource_name"
        end

        # Error check 3: Make sure category and type are consistent. For
        # example, category 'lo' and type 'EQUIP' are NOT consistent.
        if (!(type1.category.code == category.code))
          raise "the type '#{type1.code}' has category '#{type1.category.code}' " +
                "which is inconsistent with the supplied " +
                "category '#{category.code}'"
        end

#        # Error check 4: Run valid?
#        if (!resource.valid?)
#          raise "the valid? method returned false" +
#                " #{resource.errors.inspect}"
#        end
      end

      return resource
    else
      # See if there is an element with the name xml_resource_tag_name
      resource_element = root.elements[xml_resource_tag_name]
      if (resource_element.nil?)
        raise "could not find a child element '#{xml_resource_tag_name}'"
      else
        return (CRDB::Resource).from_xml(resource_element.to_s)
      end
    end
  end



#  # Returns a CRDB::Resource object (is this used anywhere???)
#  def (CRDB::Resource).create_from_xml(xml_string, *options_array)
#
#    # Extract any options and process
#    dry_run = false
#    if (options_array.length > 0)
#      options_hash = options_array[0]
#      if (options_hash.key?(:dry_run))
#        dry_run = true
#      end
#    end
#
#    resource = (CRDB::Resource).from_xml(xml_string)
#
#    # Add a resourceadmin from contact_email
#    contact_email = resource.contact_email
#
#    if (contact_email =~ /^([^@]+)@/)
#      sunetid = $1
#
#      # Convert sunetid to  Kerberos uid.
#      kerberos_ticket = (CRDB::CRDBRemctl).make_kerberos()
#      begin
#        kerb_uid = (CRDB::Misc).sunetid_to_uid(sunetid,
#                                       {:kerberos => kerberos_ticket})
#      rescue
#        # If there was an error, we assume that the contact_email
#        # is not a valid sunetid and skip the resource admin step.
#        puts "There was error gettting the kerberos id: " + $!
#        kerb_uid = nil
#      end
#
#      if (!kerb_uid.nil? && (CRDB::Misc).sunetid?(kerb_uid))
#        resourceadmin = (CRDB::Resourceadmin).new()
#        resourceadmin.sunetid = kerb_uid
#        resource.resourceadmins << resourceadmin
#      end
#    end
#
#    return resource
#  end

#  # Given a new resource, an CRDB::Adminownership, and an optional
#  # CRDB::Ownership object, create a new resource.
#  def create_new_resource()
#
#    # Construct the new uid and check that there is not already a resource
#    # with that uid.
#    new_uid = derive_uid_from_name(self.display_name)
#    raise new_uid
#
#    # Step 1. Make a fake uid and save the resource
#    time = Time.new
#    fake_uid = 'fake_uid' + Process.pid.to_s + time.to_i.to_s
#
#    self.uid = fake_uid
#    self.save!
#
#    # Step 2. Save the CRDB::Adminownership and CRDB::Ownership objects
#    if (!adminownership.nil?)
#      adminownership.save!
#    end
#
#    if (!ownership.nil?)
#      ownership.save!
#    end
#
#    # Step 3. Make a new uid based (possibly) on the ownership.
#    new_uid =  self.derive_uid_from_name(self.display_name)
#    self.uid = new_uid
#
#    # If the save fails, delete the resource.
#    begin
#      self.save!
#    rescue
#      # Delete the previously saved resource.
#      fake_resource = (CRDB::Resource).find_by_uid(fake_uid)
#
#      if (!fake_resource.nil?)
#        fake_resource.destroy
#      end
#    end
#
#  end

  # The next functions
  def (CRDB::Resource).find_resource_available_fields()
    return [
            "adminunit",
            "resourceadmin",
            "building",
            "capacity",
            "contact_email",
            "contact_fullname",
            "contact_phone",
            "department",
            "old_sundial_name",
            "type",
            "uid",
            "nickname",
            "displayname_override",
            "floor-room",
            "room_override",
            "resource_name",
            "notes",
            "policy",
           ]
  end

  # Search for all resources whose field <field_name> contains the string
  # <search_string>.
  # Avalaible options: verbose => true|false
  def (CRDB::Resource).find_resource_search_results(field_name, search_string, *options_array)

    # 0. Process options
    verbose = false
    if (options_array.length > 0)
      options_hash = options_array[0]
      if (options_hash.key?(:verbose))
        verbose = options_hash[:verbose]
      end
    end

    # 1. Both parameters are required
    if (field_name.nil?)
      raise "missing parameter: field_name"
    elsif (search_string.nil?)
      raise "missing parameter: search_string"
    end

    # 2. Check that the field_name is valid
    available_fields = find_resource_available_fields()
    if (! (available_fields.include?(field_name)))
      raise %Q['#{field_name}' is not a searchable field!]
    end

    # Change some field_name's
    if (field_name =~ /floor.room/)
      field_name = 'FR'
    end

    # 3. Replace any *'s with '%'s
    search_string_local = search_string.gsub(/\*/, "%")

    # 4. Do the search
    method_name = field_name

    if (field_name.eql?("building"))
      method_name = 'building_short_name'
      resources = (CRDB::Resource).find(:all, :include => [ :building ],
                 :conditions => ['buildings.short_name LIKE ?', "%" + search_string_local + "%"])
    elsif (field_name.eql? "department")
      method_name = 'departments_formatted'
      resources = (CRDB::Resource).find(:all, :include => [ :departments ],
                 :conditions => ['departments.code LIKE ?', "%" + search_string_local + "%"])
    elsif (field_name.eql? "resourceadmin")
      resources = (CRDB::Resource).find(:all, :include => [ :resourceadmins ],
                 :conditions => ['resourceadmins.sunetid LIKE ?', "%" + search_string_local + "%"])
    elsif (field_name.eql? "policy")
      method_name = 'policy_name'
      resources = (CRDB::Resource).find(:all, :include => [ :policy ],
                 :conditions => ['policies.name LIKE ?', "%" + search_string_local + "%"])
    elsif (field_name.eql? "type")
      method_name = 'type_code'
      resources = (CRDB::Resource).find(:all, :include => [ :type ],
                 :conditions => ['types.code LIKE ?', "%" + search_string_local + "%"])
    elsif (field_name =~ /admin_unit|adminunit/)
      method_name = 'adminunits_formatted'
      field_name = 'adminunit'
      resources = (CRDB::Resource).find(:all, :include => [ :adminunits ],
                 :conditions => ['adminunits.code LIKE ?', "%" + search_string_local + "%"])
    else
      if (field_name.eql? "capacity")
        if (search_string_local =~ /\A[1-9]+[0-9]+\z/)
          where_clause = ['resources.' + field_name + ' >= ?', search_string_local ]
        else
          raise "capacity has to be a valid digit!"
        end
      else
        where_clause = ['resources.' + field_name + ' LIKE ?', "%" + search_string_local + "%"]
      end

     # Load all resources whose contact email matches the query
     resources = (CRDB::Resource).find(:all, :conditions => where_clause )
    end

    if (resources.empty?)
      return  "no matching resources found"
    else
      return_value = ''
      resources.each do |resource|

        # We need to re-read in case we searched on a many-to-many field
        res_reread = (CRDB::Resource).find(resource.id)

        if (verbose)
          return_value += res_reread.to_s + "\n"
          return_value += "\n----------------------------"  + "\n"
        else
          return_value += "#{res_reread.uid} [#{res_reread.display_name}]" +
                          " (#{res_reread.send(method_name)})" + "\n"
        end
      end
      return return_value
    end
  end


  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###
  # "Mix-in" the Fake module
  require 'crdb/fake.rb'
  include CRDB::Fake

  @@fake_method_to_value_hash = nil
  def (CRDB::Resource).initialize_fake_method_to_value_hash()
    fake_uids = ['cal_lo_fake_abcs_qrnw_492', 'cal_eq_fake_2347324_sdfkjhdf']
    fake_contact_fullnames = ['Ferd Berfel', nil]
    fake_contact_emails = ['ferd.berfel@stanford.edu', 'resource_mgr_44219@stanford.edu']
    fake_contact_phones = ['650-555-5555', nil]

    require 'crdb/building'
    (CRDB::Building).delete_all_fakes()
    building = (CRDB::Building).new()
    fake_building0 = building.make_fake({:index => 0})
    fake_buildings = [fake_building0, nil]

    fake_FRs = ['01-111', nil]

    require 'crdb/type'
    type = (CRDB::Type).new()
    type0 = (CRDB::Type).find_by_code('CR')
    type1 = (CRDB::Type).find_by_code('CART')
    fake_types = [type0, type1]

    fake_resource_names = [nil, 'Skateboard #3']

    require 'crdb/policy'
    (CRDB::Policy).delete_all_fakes()
    policy = (CRDB::Policy).new()
    fake_policy0 = policy.make_fake({:index => 0})
    fake_policy1 = policy.make_fake({:index => 1})
    fake_policies = [fake_policy0, fake_policy1]

    @@fake_method_to_value_hash =
    {
      'uid'              => fake_uids,
      'contact_fullname' => fake_contact_fullnames,
      'contact_email'    => fake_contact_emails,
      'contact_phone'    => fake_contact_phones,
      'building'         => fake_buildings,
      'FR'               => fake_FRs,
      'type'             => fake_types,
      'resource_name'    => fake_resource_names,
      'policy'           => fake_policies,
    }
  end

  def (CRDB::Resource).get_fake_method_to_value_hash()
    return @@fake_method_to_value_hash
  end

  def (CRDB::Resource).delete_all_fakes()
    (CRDB::Resource).initialize_fake_method_to_value_hash()
    fake_uids = @@fake_method_to_value_hash['uid']
    (CRDB::Resource).destroy_all(["uid = ?", fake_uids[0]])
    (CRDB::Resource).destroy_all(["uid = ?", fake_uids[1]])
    (CRDB::Building).delete_all_fakes()
    (CRDB::Policy).delete_all_fakes()
  end
  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###  # ## ###


  # #####  # #####  # #####  # #####  # #####  # #####  # #####  # #####
  # Callbacks

  def (CRDB::Resource).before_save
    # 1. For some attributes, if the attribute is the empty string or all
    #    whitespace, replace with nil
    attributes_to_whitespace_nillify =
    [
      'contact_fullname',
      'contact_email',
      'contact_phonenumber',
      'displayname_override',
      'FR',
      'room_override',
      'nickname',
      'resource_name',
      'notes',
      'old_sundial_name',
    ]
    attributes_to_whitespace_nillify.each do |attributes_to_whitespace_nillify|
      self.send(attribute_to_whitespace_nillify + '=',
                (CRDB::Misc).ws_nillify(self.send(attribute_to_whitespace_nillify)))
    end
  end


  ## CALLBACK DEFINITIONS

  def add_event_after_create()
    require 'crdb/eventcode'
    require 'crdb/event'

    resource_created_eventcode = (CRDB::Eventcode).find_by_code('EV_RESOURCE_CREATED')
    event = (CRDB::Event).new()
    event.eventcode = resource_created_eventcode
    event.resource = self
    event.arg1 = self.uid()
    event.arg2 = self.display_name()
    event.save!
  end

  def add_event_after_update()
    require 'crdb/eventcode'
    require 'crdb/event'

    resource_created_eventcode = (CRDB::Eventcode).find_by_code('EV_RESOURCE_UPDATED')
    event = (CRDB::Event).new()
    event.eventcode = resource_created_eventcode
    event.resource = self
    changes = self.changes

    # Encode the changed fields for saving
    encoded_changes = ''
    changes.each do |field, values|
      old_value = values[0]
      new_value = values[1]

      # Skip: updated_at
      if (field =~ /updated_at/)
        next
      end

      encoded_changes += "[#{field}:(#{old_value})(#{new_value})]"
    end

    event.arg1 = encoded_changes

    event.save!
  end

  def (CRDB::Resource).dump_schema()
    return (ActiveRecord::SchemaDumper).dump() ;
  end



  end # matches "class Resource"
end # matches "module CRDB"
