require 'crdb/remctl_caladmin'
require 'crdb/resource'

module CRDB
 class Zimbra < Object

  # Define the base directory where the binaries are kept
  @bin_base_dir = '/opt/zimbra/bin/'

  # Define the zmprov and zmmailbox binary
  @ZMPROV    = @bin_base_dir + 'zmprov'
  @ZMMAILBOX = @bin_base_dir + 'zmmailbox'

  # A zimbra object must point at a particular
  # Zimbra server. The server name should be a fully-qualified
  # domain name.

  attr_accessor :server, :caching, :kerberos

  # Override the server setter
  def server=(server)
    if (!server.nil? && (server !~ /^ [^\.]+ \. stanford \. edu $/x))
      raise "server must have the format 'something.stanford.edu' " +
            "(yours is '#{server}')"
    else
      @server = server
      return @server
    end
  end

  # We want @caching to be on (true) by default.

  def caching?()
    return @caching
  end

  def caching_on()
    @caching = true
    return @caching
  end

  def caching_off()
    @caching = false
    return @caching
  end

  def parse_truefalse(string)
    if (string =~ /true/i)
      return true
    elsif (string =~ /false/i)
      return false
    else
      raise "do not know what '#{string}' is"
    end
  end

  # Gets all the resources on the Zimbra server.
  # Returns a hash mapping zimbra_name address to CRDB::Resource object.
  def get_zimbra_name_to_resource()
    ## STAGE 1. Get all the attributes from the server
    remctl = (CRDB::RemctlCaladmin).new()
    remctl.server = @server
    if (!@kerberos.nil?)
       remctl.kerberos = @kerberos
    end
    # remctl.debug_on()

    # If caching is turned on, turn on the remctl's caching. Otherwise,
    # turn the remctl caching off and clear it.
    if (@caching)
      remctl.caching_on()
    else
      remctl.clear_cache()
      remctl.caching_off()
    end

    # Dump the attributes
    dump_attributes = remctl.dump_attributes()

    # Now we parse the attributes creating a new resources object for each
    # section
    zimbra_name_to_resource = parse_attributes(dump_attributes)

    return zimbra_name_to_resource
  end

  # Given an attribute dump from a remctl call, create resources from it.
  # Returns a hash mapping Zimbra name to a CRDB::Resource object.
  def parse_attributes(attribute_string)
    zimbra_name_to_resource = {}

    lines = attribute_string.split("\n")
    resource = nil

    # Attributes lines should look like this:
    #
    # # name cal_sc_holiday@stanford.edu
    # cn: Holiday Calendar
    # displayName: Holiday Calendar
    # mail: cal_sc_holiday@stanford.edu
    # objectClass: organizationalPerson
    # objectClass: zimbraAccount
    # objectClass: amavisAccount
    # objectClass: zimbraCalendarResource
    # etc...

    # The first line (the one with '# name ...') tells us the uid and
    # domain.

    lines.each do |line|
      if (line =~ /^ \# \s* name \s* (\S+)(\@)(stanford.edu) \s* $/x)
        uid    = $1
        domain = $3
        resource = (CRDB::Resource).new()
        resource.uid = uid
        resource.domain = domain
        zimbra_name_to_resource[resource.zimbra_name] = resource
      elsif (line =~ /^ displayName\: \s* (.+) \s* /x)
        # This is the displayName line
        displayName = $1
        displayName.strip!
        resource.za_displayName = displayName
        resource.displayname_override = displayName
      elsif (line =~ /^ description\: \s* (.+) \s* /x)
        description = $1.strip
        resource.za_description = description
      elsif (line =~ /^ zimbraNotes\: \s* (.+) \s* /x)
        resource.za_zimbraNotes = $1.strip
      elsif (line =~ /^ zimbraCalResContactEmail\: \s* (\S+) \s* /x)
        # This is a Contact Email line
        resource.contact_email = $1
      elsif (line =~ /^ zimbraCalResContactName\: \s* (.*) \s* /x)
        # This is a Contact Name line
        resource.contact_fullname = $1
      elsif (line =~ /^ zimbraCalResContactPhone\: \s* (.*) \s* /x)
        # This is a Contact Phone line
        resource.contact_phone = $1
      elsif (line =~ /^ zimbraCalResCapacity\: \s* (.*) \s* /x)
        # This is a Capacity line
        resource.capacity = $1
      elsif (line =~ /^ zimbraCalResSite\: \s* (.*) \s* /x)
        # This is a Site line
        resource.za_zimbraCalResSite = $1
      elsif (line =~ /^ zimbraCalResBuilding\: \s* (.*) \s* /x)
        # This is a Building line
        building_name = $1
        resource.za_zimbraCalResBuilding = building_name
        # Try to get the building id
        building = (CRDB::Building).find_by_very_short_name(building_name)
        if (!building.nil?)
          resource.building = building
        else
          # Nothing
        end
      elsif (line =~ /^ zimbraCalResFloor\: \s* (.*) \s* /x)
        # This is a Floor line
        resource.za_zimbraCalResFloor = $1
      elsif (line =~ /^ zimbraCalResRoom\: \s* (.*) \s* /x)
        # This is a Room line
        resource.za_zimbraCalResRoom = $1
        resource.room_override = $1
      elsif (line =~ /^ zimbraCalResLocationDisplayName\: \s* (.*) \s* /x)
        # This is the Location display line that appears in resource
        # search results.
        resource.za_zimbraCalResLocationDisplayName = $1
      elsif (line =~ /^ zimbraCalResAutoAcceptDecline\: \s* (.*) \s* /x)
        # This is a zimbraCalResAutoAcceptDecline line
        resource.za_zimbraCalResAutoAcceptDecline = parse_truefalse($1)
      elsif (line =~ /^ zimbraCalResAutoDeclineIfBusy\: \s* (.*) \s* /x)
        # This is a zimbraCalResAutoDeclineIfBusy line
        resource.za_zimbraCalResAutoDeclineIfBusy = parse_truefalse($1)
      elsif (line =~ /^ zimbraCalResAutoDeclineRecurring\: \s* (.*) \s* /x)
        # This is a zimbraCalResAutoDeclineRecurring line
        resource.za_zimbraCalResAutoDeclineRecurring = parse_truefalse($1)
      elsif (line =~ /^ zimbraCalResType\: \s* (.*) \s* /x)
        # This is a zimbraCalResType line
        resource_type = $1
        if (resource_type !~ /equipment|location/i)
          raise "found unknown zimbraCalResType: $1"
        else
          resource.za_zimbraCalResType = resource_type
        end
      elsif (line =~ /^ mail\: \s* (cal_[^\@]+\@stanford\.edu)\s* /x)
        # This is the e-mail?
      else
        # We don't care about this line.
      end
    end

    # Loop through the resources and set the policy based on the
    # zimbraCalRes attributes.
    zimbra_name_to_resource.each do |zimbra_name, resource|
      resource.derive_policy_from_attributes()
    end

    return zimbra_name_to_resource
  end

  # Make a resource object from a Zimbra resource.
  def get_resource(zimbra_name)
    # Get this resource's attribute from Zimbra.
    remctl = (CRDB::RemctlCaladmin).new()
    remctl.server = @server
    remctl.caching_off()
    remctl.kerberos = @kerberos

    begin
      attributes = remctl.get_all_attributes(zimbra_name)
    rescue RuntimeError
      if ($! =~ /access.denied/i)
        raise "there was an error running the remctl command: " + $!
      else
        raise "the resource #{zimbra_name} is not on #{self.server}: " + $!
      end
    end

    # Parse the attributes.
    zimbra_name_to_resource = parse_attributes(attributes)

    if (zimbra_name_to_resource.length <= 0)
      raise "no resources found for '#{zimbra_name}'"
    elsif (zimbra_name_to_resource.length > 1)
      raise "too may resources found for '#{zimbra_name}'"
    else
      resource = zimbra_name_to_resource.to_a[0][1]
      return resource
    end
  end


  # Given a CRDB::Resource object, push the objects attributes up to the Zimbra
  # server. If the attribute does not exist, raise an exception.
  #
  # The first argument should be a CRDB::Resource object.
  # The second argument can either be nil, which implies that all
  # attributes should be synced, or an array of Zimbra attributes to
  def sync_resource(database_resource, attributes_to_sync_array, *options_array)

    ###
    ### STEP 1. Extract any options and process
    dry_run = false
    if (options_array.length > 0)
      options_hash = options_array[0]
      if (options_hash.key?(:dry_run))
        dry_run = options_hash[:dry_run]
      end
    end


    ###
    ### STEP 2. Sync all the attributes (except resource admins which we
    ### handle in STEP 3).
    # Get the resource's zimbra_name.
    zimbra_name = database_resource.zimbra_name

    # Keep track of the number of changes needed/made
    number_of_changes = 0

    # Get the CRDB::Resource object on Zimbra corresponding to this uid.
    zimbra_resource = self.get_resource(zimbra_name)
    if (zimbra_resource.nil?)
      raise "problem finding Zimbra resource for '#{zimbra_name}'"
    end


    remctl = (CRDB::RemctlCaladmin).new()
    remctl.server = @server
    remctl.caching_off()
    remctl.kerberos = @kerberos

    # This is a hash that includes all the attributes we want
    # to update on the Zimbra server.
    zimbra_attributes_to_method =
    {
      'displayName'              => 'display_name',
      'zimbraCalResContactName'  => 'contact_fullname',
      'zimbraCalResContactEmail' => 'contact_email',
      'zimbraCalResContactPhone' => 'contact_phone',
      'description'              => 'description',
      'zimbraCalResBuilding'     => 'building_very_short_name',
      'zimbraCalResRoom'         => 'room',
      'zimbraCalResCapacity'     => 'capacity',
      'zimbraCalResLocationDisplayName' => 'zimbraLocationDisplayName',
      'zimbraCalResType'         => 'zimbraCalResType',
    }

    return_value = ''

    ## ###      ## ###      ## ###      ## ###      ## ###      ## ###      ## ###
    # Define a little Proc that sends some output to the screen
    # informing the user what is happening.
    progress = Proc.new do |meth, zim_val, db_val|
      if (dry_run)
        return_value += "would have changed Zimbra #{meth}:" + "\n"
      else
        return_value += "changing Zimbra #{meth}:"
      end
      zim_val_q = (CRDB::Misc).what_is_this_string(zim_val)
      db_val_q  = (CRDB::Misc).what_is_this_string(db_val)
      return_value += " FROM #{zim_val_q}\n" +
                      " TO #{db_val_q}" + "\n"
    end
    ## ###      ## ###      ## ###      ## ###      ## ###      ## ###      ## ###

    # Start updating.
    zimbra_attributes_to_method.each do |attribute, method|
      # Skip this attribute if it does NOT appear in attributes_to_sync_array.
      if ( !attributes_to_sync_array.nil? &&
          (!attributes_to_sync_array.include?(attribute))
         )
        next
      end

      database_value = database_resource.send(method)
      zimbra_value   = zimbra_resource.send(method)

      if (database_value != zimbra_value)
        number_of_changes += 1

        progress.call(method, zimbra_value, database_value)
        if (!dry_run)
          remctl.set_attribute(zimbra_name, attribute, database_value)
        end
      end
    end

    # We handle policy differences separately.
    if (!database_resource.policy.eql?(zimbra_resource.policy))
      number_of_changes += 1
      progress.call('policy', zimbra_resource.policy.name, database_resource.policy.name)
      if (!dry_run)
        remctl.set_policy(zimbra_name, database_resource.policy.name)
      end
    end


    ###
    ### STEP 3. We need to grant to all the resourceadmins admin rights
    number_resource_admins_added = 0

    # Get the permissions for the '/Calendar' folder
    folderCalendar = (CRDB::ZimbraFolder).new()
    folderCalendar.resource = database_resource
    folderCalendar.path = '/Calendar'
    folderCalendar.server = @server
    folderCalendar.kerberos = @kerberos
    folderCalendar.populate_permissions()
    permissions = folderCalendar.permissions

    resourceadmins = database_resource.resourceadmins

    # Any sunetids in resourceadmins need to be added as resource
    # admins in Zimbra.
    resourceadmins.each do |resourceadmin|

      sunetid = resourceadmin.sunetid
      if (!sunetid.nil?)
        # Grant this sunetid admin rights, but only if it is not already
        # an administrator.

        # Get the uid from the sunetid
        sunetid_uid = (CRDB::Misc).sunetid_to_uid(sunetid, {:kerberos => @kerberos})

        if (!sunetid_uid.eql?(sunetid))
          return_value += "Warning: the resource admin '#{sunetid}' does " +
                          "not match the Kerberos uid '#{sunetid_uid}'; " +
                          "please fix." + "\n"
        end

        uid_email = sunetid_uid + '@stanford.edu'
        if (permissions.key?(uid_email))
          # This user is in the list. Is it an admin?
          permission = permissions[uid_email]
          if (permission.can_administer)
            # YES! This sunetid ALREADY can administer, so do not grant.
            return_value += "#{sunetid} already has administer rights"
            grant_rights = false
          else
            # NO! This sunetid does NOT have administer rights, so grant.
            return_value += "#{sunetid} has some rights but not administer rights"
            grant_rights = true
          end
        else
          # This sunetid does not have any current rights, so grant.
            return_value += "#{sunetid} currently has no rights"
            grant_rights = true
        end

        if (grant_rights)
          if (dry_run)
            return_value += ": would have added admin rights for user #{sunetid}" + "\n"
          else
            return_value += ": adding admin rights for user #{sunetid}" + "\n"
            remctl.share(database_resource.uid, sunetid_uid, 'admin')
          end
          number_resource_admins_added += 1
        else
          return_value += ": not adding admin rights for user #{sunetid}" + "\n"
        end

      end
    end

#    # Step 3.2 Any users in the permissions list that have admin rights
#    # but are NOT in the resourceadmins array need to be removed as resource admin
#    # in Zimbra.
#    permissions.each do |principal, permission|
#      # Does this principal have administer rights?
#      if (permission.can_administer)
#        # Yes, they can. See if this principal is a resource admin.
#        # If not, provide a warning.
#        if (principal =~ /^([^@]+)@stanford.edu$/)
#          sunetid = $1
#        end
#
#
#      end
#    end


    if (number_of_changes == 0)
      return_value += "No attribute changes needed." + "\n"
    end

    if (number_resource_admins_added == 0)
       return_value += "No resource admins added." + "\n"
    end

    # [17-Aug-2009] Even if nothing was changed, we still
    # update the last_sync_date field.
    database_resource.last_sync_time = Time.now
    database_resource.save

    return return_value
  end


end

class CRDB::ZimbraFolderPermission < Object

  # Permissions    Type  Display
  # -----------  ------  -------
  #      rwidxa  accoun  yuelu@stanford.edu
  #       rwidx  accoun  tinag@stanford.edu
  #       rwidx  accoun  luvien@stanford.edu
  #       rwidx  accoun  cholada@stanford.edu
  #       rwidx  accoun  ptyoung@stanford.edu
  #       rwidx  accoun  somjai@stanford.edu
  #       rwidx  accoun  ginar@stanford.edu
  #       rwidx  accoun  mcclish@stanford.edu
  #       rwidx  accoun  annap@stanford.edu
  #       rwidx  accoun  ebuni@stanford.edu
  #       rwidx  accoun  rajsingh@stanford.edu
  #       rwidx  accoun  lkimbosw@stanford.edu
  #       rwidx  accoun  jrocha@stanford.edu
  #           r     all
  #       rwidx  accoun  adamhl@stanford.edu
  #       rwidx  accoun  jcowart@stanford.edu

  # r - read
  # w - write
  # i - insert
  # d - delete
  # a - administer
  # x - accept/decline invites

  attr_accessor :type  # The principal's type (account, all, etc.)
  attr_accessor :uid   # The account's uid
  attr_accessor :can_read
  attr_accessor :can_write
  attr_accessor :can_insert
  attr_accessor :can_delete
  attr_accessor :can_accept_decline_invites
  attr_accessor :can_administer

  def initialize()
    @can_read = false
    @can_write = false
    @can_insert = false
    @can_delete = false
    @can_accept_decline_invites = false
    @can_administer = false
    @can_access_private = false
  end

  # Permissions    Type  Display
  # -----------  ------  -------
  #      rwidxa  accoun  yuelu@stanford.edu
  #       rwidx  accoun  tinag@stanford.edu
  def to_s
    rv = ''

    # permissions string
    perm_string = self.to_permission_string()
    spacer = ' ' * (11 - perm_string.length)
    rv += spacer + perm_string

    spacer = ' ' * (8 - self.type.length)
    rv += spacer + self.type

    spacer = ' ' * 2
    rv += spacer + "#{self.uid}"

    return rv
  end


  # Parse a string of the form "rwi.."
  def from_permission_string(perm_string)
    if (perm_string =~ /r/)
      @can_read = true
    end

    if (perm_string =~ /w/)
      @can_write = true
    end

    if (perm_string =~ /i/)
      @can_insert = true
    end

    if (perm_string =~ /d/)
      @can_delete = true
    end

    if (perm_string =~ /x/)
      @can_accept_decline_invites = true
    end

    if (perm_string =~ /a/)
      @can_administer = true
    end

    if (perm_string =~ /p/)
      @can_access_private = true
    end

    return
  end

  # Convert to a permissions string
  def to_permission_string()
    rv = ''

    if (@can_read)
      rv += 'r'
    end

    if (@can_write)
      rv += 'w'
    end

    if (@can_insert)
      rv += 'i'
    end

    if (@can_delete)
      rv += 'd'
    end

    if (@can_accept_decline_invites)
      rv += 'x'
    end

    if (@can_administer)
      rv += 'a'
    end

    if (@can_access_private)
      rv += 'p'
    end

    return rv
  end

  # Given a Zimbra uid and domain, create a hash mapping
  # Zimbra principals to a CRDB::ZimbraFolderPermission object.
  def (CRDB::ZimbraFolderPermission).populate(uid, server, domain = 'stanford.edu',  *options_array)

    # Extract any options and process
    kerberos = nil
    if (options_array.length > 0)
      options_hash = options_array[0]
      if (options_hash.key?(:kerberos))
        kerberos = options_hash[:kerberos]
      end
    end

    remctl = (CRDB::RemctlCaladmin).new()
    remctl.server = server
    remctl.caching_off()
    if (!kerberos.nil?)
      remctl.kerberos = kerberos
    end

    listacl_string = remctl.listacl(uid)

    # Parse the listacl_string creating permissions
    permissions = {}
    lines = listacl_string.split("\n")
    lines.each  do |line|
      # rwidxa  accoun  yuelu@stanford.edu
      if (line =~ /^\s+([rwidxap]\S*)\s+(\S+)\s+($|\S.*$)/)
        perms = $1
        type  = $2
        uid   = $3
        perm_object = (CRDB::ZimbraFolderPermission).new()
        perm_object.type = type.strip
        perm_object.uid  = uid.strip
        perm_object.from_permission_string(perms)
        permissions[uid] = perm_object
      end
    end

    return permissions
  end

end

class CRDB::ZimbraFolder < CRDB::Zimbra

  # Define some properties/accessors
  attr_accessor :resource    # The CRDB::Resource object that "owns" this folder.
  attr_accessor :path        # The path of this folder (e.g., "/Calendar")
                             # Note that the path INCLUDES the forward slash.
  attr_accessor :permissions # A hash mapping Zimbra permission principals
                             # to CRDB::ZimbraFolderPermission objects

  def to_s
    resource_formatted = "#{@resource.uid} [#{@resource.display_name}]"
    path_formatted     = @path

    if (@permissions.nil?)
      permissions_formatted = 'NO PERMISSIONS (did you forget to populate permissions?)'
    else
      permissions_formatted = ''
      permission_principals = @permissions.keys.sort()
      permission_principals.each do |principal|
        permission = @permissions[principal]
        permissions_formatted += permission.to_s + "\n"
      end
    end

    return resource_formatted + "\n" +
           path_formatted + "\n" +
           permissions_formatted
  end

  # Populate the permissions attribute
  def populate_permissions()
    @permissions = (CRDB::ZimbraFolderPermission).populate(@resource.uid, @server, 'stanford.edu',
                                                         {:kerberos => @kerberos})
  end

 end # matches "module CRDB"
end # matches "class Zimbra"
